"""
input/output streams.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/io.rst.
=================================

.. module:: io
   :synopsis: input/output streams

|see_cpython_module| :mod:`python:io`.

This module contains additional types of `stream` (file-like) objects
and helper functions.

Conceptual hierarchy
--------------------

.. admonition:: Difference to CPython
   :class: attention

   Conceptual hierarchy of stream base classes is simplified in MicroPython,
   as described in this section.

(Abstract) base stream classes, which serve as a foundation for behaviour
of all the concrete classes, adhere to few dichotomies (pair-wise
classifications) in CPython. In MicroPython, they are somewhat simplified
and made implicit to achieve higher efficiencies and save resources.

An important dichotomy in CPython is unbuffered vs buffered streams. In
MicroPython, all streams are currently unbuffered. This is because all
modern OSes, and even many RTOSes and filesystem drivers already perform
buffering on their side. Adding another layer of buffering is counter-
productive (an issue known as "bufferbloat") and takes precious memory.
Note that there still cases where buffering may be useful, so we may
introduce optional buffering support at a later time.

But in CPython, another important dichotomy is tied with "bufferedness" -
it's whether a stream may incur short read/writes or not. A short read
is when a user asks e.g. 10 bytes from a stream, but gets less, similarly
for writes. In CPython, unbuffered streams are automatically short
operation susceptible, while buffered are guarantee against them. The
no short read/writes is an important trait, as it allows to develop
more concise and efficient programs - something which is highly desirable
for MicroPython. So, while MicroPython doesn't support buffered streams,
it still provides for no-short-operations streams. Whether there will
be short operations or not depends on each particular class' needs, but
developers are strongly advised to favour no-short-operations behaviour
for the reasons stated above. For example, MicroPython sockets are
guaranteed to avoid short read/writes. Actually, at this time, there is
no example of a short-operations stream class in the core, and one would
be a port-specific class, where such a need is governed by hardware
peculiarities.

The no-short-operations behaviour gets tricky in case of non-blocking
streams, blocking vs non-blocking behaviour being another CPython dichotomy,
fully supported by MicroPython. Non-blocking streams never wait for
data either to arrive or be written - they read/write whatever possible,
or signal lack of data (or ability to write data). Clearly, this conflicts
with "no-short-operations" policy, and indeed, a case of non-blocking
buffered (and this no-short-ops) streams is convoluted in CPython - in
some places, such combination is prohibited, in some it's undefined or
just not documented, in some cases it raises verbose exceptions. The
matter is much simpler in MicroPython: non-blocking stream are important
for efficient asynchronous operations, so this property prevails on
the "no-short-ops" one. So, while blocking streams will avoid short
reads/writes whenever possible (the only case to get a short read is
if end of file is reached, or in case of error (but errors don't
return short data, but raise exceptions)), non-blocking streams may
produce short data to avoid blocking the operation.

The final dichotomy is binary vs text streams. MicroPython of course
supports these, but while in CPython text streams are inherently
buffered, they aren't in MicroPython. (Indeed, that's one of the cases
for which we may introduce buffering support.)

Note that for efficiency, MicroPython doesn't provide abstract base
classes corresponding to the hierarchy above, and it's not possible
to implement, or subclass, a stream class in pure Python.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from types import TracebackType
from typing import TypeVar, Final, Protocol, runtime_checkable, Literal
from typing import AnyStr, overload, Type

from uarray import array

_T: Final = TypeVar("_T")

_OpenTextModeUpdating: Final = Literal[
    "r+",
    "+r",
    "rt+",
    "r+t",
    "+rt",
    "tr+",
    "t+r",
    "+tr",
    "w+",
    "+w",
    "wt+",
    "w+t",
    "+wt",
    "tw+",
    "t+w",
    "+tw",
    "a+",
    "+a",
    "at+",
    "a+t",
    "+at",
    "ta+",
    "t+a",
    "+ta",
    "x+",
    "+x",
    "xt+",
    "x+t",
    "+xt",
    "tx+",
    "t+x",
    "+tx",
]
_OpenTextModeWriting: Final = Literal["w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"]
_OpenTextModeReading: Final = Literal[
    "r", "rt", "tr", "U", "rU", "Ur", "rtU", "rUt", "Urt", "trU", "tUr", "Utr"
]
_OpenTextMode: Final = _OpenTextModeUpdating | _OpenTextModeWriting | _OpenTextModeReading

_OpenBinaryModeUpdating: Final = Literal[
    "rb+",
    "r+b",
    "+rb",
    "br+",
    "b+r",
    "+br",
    "wb+",
    "w+b",
    "+wb",
    "bw+",
    "b+w",
    "+bw",
    "ab+",
    "a+b",
    "+ab",
    "ba+",
    "b+a",
    "+ba",
    "xb+",
    "x+b",
    "+xb",
    "bx+",
    "b+x",
    "+bx",
]
_OpenBinaryModeWriting: Final = Literal["wb", "bw", "ab", "ba", "xb", "bx"]
_OpenBinaryModeReading: Final = Literal[
    "rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"
]
_OpenBinaryMode: Final = _OpenBinaryModeUpdating | _OpenBinaryModeReading | _OpenBinaryModeWriting

AnyStr_co: Final = TypeVar("AnyStr_co", str, bytes, covariant=True)
@runtime_checkable
class PathLike(Protocol[AnyStr_co]):
    def __fspath__(self) -> AnyStr_co: ...

StrOrBytesPath: Final = str | bytes | PathLike[str] | PathLike[bytes]
_OpenFile: Final = StrOrBytesPath | int

AnyReadableBuf: Final = TypeVar("AnyReadableBuf", bytearray, array, memoryview, bytes)
"""
Type that allows bytearray, array, memoryview, or bytes, 
but only one of these and not a mixture in a single declaration.
"""

AnyWritableBuf: Final = TypeVar("AnyWritableBuf", bytearray, array, memoryview)
"""
Type that allows bytearray, array, or memoryview, but only one of these and not a mixture in a single declaration.
"""

_Self: Final = TypeVar("_Self")  # The type that extends `IOBase`.
@runtime_checkable
class IOBase(Protocol[AnyStr, _Self]):
    """A `Protocol` (structurally typed) for an IOStream."""

    __slots__ = ()
    def __enter__(self) -> _Self:
        """
        Called on entry to a `with` block.
        The `with` statement will bind this method’s return value to the target(s) specified in the `as` clause 
        of the statement, if any.
        """
    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """
        Called on exit of a `with` block.
        The parameters describe the exception that caused the context to be exited. 
        If the context was exited without an exception, all three arguments will be `None`.

        If an exception is supplied, and the method wishes to suppress the exception 
        (i.e., prevent it from being propagated), it should return a true value. 
        Otherwise, the exception will be processed normally upon exit from this method.

        *Note* that `__exit__()` methods should not re-raise the passed-in exception; 
        this is the caller’s responsibility.
        """
    def __next__(self) -> AnyStr:
        """
        Next string.
        """
    def __iter__(self) -> _Self:
        """
        Start new iteration.
        """
    def close(self) -> None:
        """
        Flushes the write buffers and closes the IO stream; best not called directly, use a `with` block instead.
        Calling `f.close()` without using a `with` block might result in content not being completely written to the 
        disk, even if the program exits successfully.
        A closed file cannot be read or written any more. 
        Any operation which requires that the file be open will raise a `ValueError` after the file has been closed. 
        Calling `f.close()` more than once is allowed.
        """
    def flush(self) -> None:
        """
        Flushes the write buffers of the IO stream.
        `flush()` does not necessarily write the file’s data to disk. 
        Use `f.flush()` followed by `os.sync()` to ensure this behavior.
        
        This method does nothing for read-only and non-blocking streams.
        """
    def read(self, size: int | None = -1) -> AnyStr | None:
        """
        Read up to `size` bytes from the object and return them as a `str` (text file) or `bytes` (binary file). 
        As a convenience, if `size` is unspecified or -1, all bytes until EOF are returned. 
        Otherwise, only one system call is ever made. 
        Fewer than `size` bytes may be returned if the operating system call returns fewer than `size` bytes.

        If 0 bytes are returned, and `size` was not 0, this indicates end of file. 
        If `self` is in non-blocking mode and no bytes are available, `None` is returned.
        """
    def readinto(self, b: AnyWritableBuf) -> int | None:
        """
        Read bytes into a pre-allocated, writable bytes-like object b, and return the number of bytes read. 
        For example, b might be a bytearray. 
        
        If `self` is in non-blocking mode and no bytes are available, `None` is returned.
        """
    def readline(self, size: int = -1) -> AnyStr:
        """
        Read and return, as a `str` (text file) or `bytes` (binary file), one line from the stream. 
        If size is specified, at most size bytes will be read.
        
        The line terminator is always `b'
'` for binary files; 
        for text files, the newline argument to `open()` can be used to select the line terminator(s) recognized.
        """
    def readlines(self, hint: int | None = -1) -> list[AnyStr]:
        """
        Read and return a list of lines, as a `list[str]` (text file) or `list[bytes]` (binary file), from the stream. 
        `hint` can be specified to control the number of lines read: 
        no more lines will be read if the total size (in bytes/characters) of all lines so far exceeds `hint`.

        `hint` values of 0 or less, as well as `None`, are treated as no hint.
        The line terminator is always `b'
'` for binary files; 
        for text files, the newline argument to `open()` can be used to select the line terminator(s) recognized.

        *Note* that it’s already possible to iterate on file objects using `for line in file: ...` 
        without calling `file.readlines()`.
        """
    def write(self, b: AnyReadableBuf) -> int | None:
        """
        Write the given bytes-like object, `b`, to the underlying raw stream, and return the number of bytes written. 
        This can be less than the length of `b` in bytes, depending on specifics of the underlying raw stream, 
        and especially if it is in non-blocking mode. 
        `None` is returned if the raw stream is set not to block and no single byte could be readily written to it. 
        
        The caller may release or mutate `b` after this method returns, 
        so the implementation only access `b` during the method call.
        """
    def seek(self, offset: int, whence: int = 0) -> int:
        """
        Change the stream position to the given byte `offset`. 
        `offset` is interpreted relative to the position indicated by `whence`.
        The default value for whence is 0. 
        
        Values for whence are:

          * 0 – start of the stream (the default); offset should be zero or positive.
          * 1 – current stream position; offset may be negative.
          * 2 – end of the stream; offset is usually negative.
        
        Returns the new absolute position.
        """
    def tell(self) -> int:
        """
        Return the current stream position.
        """

@overload
def open(name: _OpenFile, /, **kwargs) -> "TextIOWrapper":
    """
    Open a file. Builtin ``open()`` function is aliased to this function.
    All ports (which provide access to file system) are required to support
    *mode* parameter, but support for other arguments vary by port.
   """

@overload
def open(name: _OpenFile, mode: _OpenTextMode = ..., /, **kwargs) -> "TextIOWrapper":
    """
    Open a file. Builtin ``open()`` function is aliased to this function.
    All ports (which provide access to file system) are required to support
    *mode* parameter, but support for other arguments vary by port.
   """

@overload
def open(name: _OpenFile, mode: _OpenBinaryMode = ..., /, **kwargs) -> "FileIO":
    """
    Open a file. Builtin ``open()`` function is aliased to this function.
    All ports (which provide access to file system) are required to support
    *mode* parameter, but support for other arguments vary by port.
   """

class FileIO(IOBase[bytes, "FileIO"]):
    """
Bytes stream from a file.
   """

    def __init__(self, name: _OpenFile, mode: str = ..., /, **kwargs):
        """
       This is type of a file open in binary mode, e.g. using ``open(name, "rb")``.
       You should not instantiate this class directly.
      """

class TextIOWrapper(IOBase[str, "TextIOWrapper"]):
    """
Str stream from a file.
   """

    def __init__(self, name: _OpenFile, mode: str = ..., /, **kwargs):
        """
       This is type of a file open in text mode, e.g. using ``open(name, "rt")``.
       You should not instantiate this class directly.
      """

class StringIO(IOBase[str, "StringIO"]):
    """
Str stream from a str (wrapper).
   """

    @overload
    def __init__(self, string: str = "", /):
        """
   
   In-memory file-like object for input/output.
   `StringIO` is used for text-mode I/O (similar to a normal file opened with "t" modifier).
   Initial contents can be specified with `string` parameter.
   
   `alloc_size` constructor creates an empty `StringIO` object, 
   pre-allocated to hold up to `alloc_size` number of bytes. 
   That means that writing that amount of bytes won't lead to reallocation of the buffer, 
   and thus won't hit out-of-memory situation or lead to memory fragmentation. 
   This constructor is a MicroPython extension and is recommended for usage only in special
   cases and in system-level libraries, not for end-user applications.

     .. admonition:: Difference to CPython
        :class: attention

        This constructor is a MicroPython extension.
      """
    @overload
    def __init__(self, alloc_size: int, /):
        """
   
   In-memory file-like object for input/output.
   `StringIO` is used for text-mode I/O (similar to a normal file opened with "t" modifier).
   Initial contents can be specified with `string` parameter.
   
   `alloc_size` constructor creates an empty `StringIO` object, 
   pre-allocated to hold up to `alloc_size` number of bytes. 
   That means that writing that amount of bytes won't lead to reallocation of the buffer, 
   and thus won't hit out-of-memory situation or lead to memory fragmentation. 
   This constructor is a MicroPython extension and is recommended for usage only in special
   cases and in system-level libraries, not for end-user applications.

     .. admonition:: Difference to CPython
        :class: attention

        This constructor is a MicroPython extension.
      """
    def getvalue(self) -> str:
        """Get the current contents of the underlying buffer which holds data."""

class BytesIO(IOBase[bytes, "BytesIO"]):
    """
Bytes stream from a bytes array (wrapper).
   """

    @overload
    def __init__(self, string: bytes = "", /):
        """
       In-memory file-like objects for input/output. `StringIO` is used for
       text-mode I/O (similar to a normal file opened with "t" modifier).
       `BytesIO` is used for binary-mode I/O (similar to a normal file
       opened with "b" modifier). Initial contents of file-like objects
       can be specified with *string* parameter (should be normal string
       for `StringIO` or bytes object for `BytesIO`). All the usual file
       methods like ``read()``, ``write()``, ``seek()``, ``flush()``,
       ``close()`` are available on these objects, and additionally, a
       following method:
       
   
   `alloc_size` constructor creates an empty `BytesIO` object, 
   pre-allocated to hold up to `alloc_size` number of bytes. 
   That means that writing that amount of bytes won't lead to reallocation of the buffer, 
   and thus won't hit out-of-memory situation or lead to memory fragmentation. 
   This constructor is a MicroPython extension and is recommended for usage only in special
   cases and in system-level libraries, not for end-user applications.

     .. admonition:: Difference to CPython
        :class: attention

        This constructor is a MicroPython extension.
      """
    @overload
    def __init__(self, alloc_size: int, /):
        """
       In-memory file-like objects for input/output. `StringIO` is used for
       text-mode I/O (similar to a normal file opened with "t" modifier).
       `BytesIO` is used for binary-mode I/O (similar to a normal file
       opened with "b" modifier). Initial contents of file-like objects
       can be specified with *string* parameter (should be normal string
       for `StringIO` or bytes object for `BytesIO`). All the usual file
       methods like ``read()``, ``write()``, ``seek()``, ``flush()``,
       ``close()`` are available on these objects, and additionally, a
       following method:
       
   
   `alloc_size` constructor creates an empty `BytesIO` object, 
   pre-allocated to hold up to `alloc_size` number of bytes. 
   That means that writing that amount of bytes won't lead to reallocation of the buffer, 
   and thus won't hit out-of-memory situation or lead to memory fragmentation. 
   This constructor is a MicroPython extension and is recommended for usage only in special
   cases and in system-level libraries, not for end-user applications.

     .. admonition:: Difference to CPython
        :class: attention

        This constructor is a MicroPython extension.
      """
    def getvalue(self) -> bytes:
        """
           Get the current contents of the underlying buffer which holds data.
      """



"""
access binary data in a structured way.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/uctypes.rst.
========================================================

.. module:: uctypes
   :synopsis: access binary data in a structured way

This module implements "foreign data interface" for MicroPython. The idea
behind it is similar to CPython's ``ctypes`` modules, but the actual API is
different, streamlined and optimized for small size. The basic idea of the
module is to define data structure layout with about the same power as the
C language allows, and then access it using familiar dot-syntax to reference
sub-fields.

.. warning::

    ``uctypes`` module allows access to arbitrary memory addresses of the
    machine (including I/O and control registers). Uncareful usage of it
    may lead to crashes, data loss, and even hardware malfunction.

.. seealso::

    Module :mod:`struct`
        Standard Python way to access binary data structures (doesn't scale
        well to large and complex structures).

Usage examples::

    import uctypes

    # Example 1: Subset of ELF file header
    # https://wikipedia.org/wiki/Executable_and_Linkable_Format#File_header
    ELF_HEADER = {
        "EI_MAG": (0x0 | uctypes.ARRAY, 4 | uctypes.UINT8),
        "EI_DATA": 0x5 | uctypes.UINT8,
        "e_machine": 0x12 | uctypes.UINT16,
    }

    # "f" is an ELF file opened in binary mode
    buf = f.read(uctypes.sizeof(ELF_HEADER, uctypes.LITTLE_ENDIAN))
    header = uctypes.struct(uctypes.addressof(buf), ELF_HEADER, uctypes.LITTLE_ENDIAN)
    assert header.EI_MAG == b"\x7fELF"
    assert header.EI_DATA == 1, "Oops, wrong endianness. Could retry with uctypes.BIG_ENDIAN."
    print("machine:", hex(header.e_machine))


    # Example 2: In-memory data structure, with pointers
    COORD = {
        "x": 0 | uctypes.FLOAT32,
        "y": 4 | uctypes.FLOAT32,
    }

    STRUCT1 = {
        "data1": 0 | uctypes.UINT8,
        "data2": 4 | uctypes.UINT32,
        "ptr": (8 | uctypes.PTR, COORD),
    }

    # Suppose you have address of a structure of type STRUCT1 in "addr"
    # uctypes.NATIVE is optional (used by default)
    struct1 = uctypes.struct(addr, STRUCT1, uctypes.NATIVE)
    print("x:", struct1.ptr[0].x)


    # Example 3: Access to CPU registers. Subset of STM32F4xx WWDG block
    WWDG_LAYOUT = {
        "WWDG_CR": (0, {
            # BFUINT32 here means size of the WWDG_CR register
            "WDGA": 7 << uctypes.BF_POS | 1 << uctypes.BF_LEN | uctypes.BFUINT32,
            "T": 0 << uctypes.BF_POS | 7 << uctypes.BF_LEN | uctypes.BFUINT32,
        }),
        "WWDG_CFR": (4, {
            "EWI": 9 << uctypes.BF_POS | 1 << uctypes.BF_LEN | uctypes.BFUINT32,
            "WDGTB": 7 << uctypes.BF_POS | 2 << uctypes.BF_LEN | uctypes.BFUINT32,
            "W": 0 << uctypes.BF_POS | 7 << uctypes.BF_LEN | uctypes.BFUINT32,
        }),
    }

    WWDG = uctypes.struct(0x40002c00, WWDG_LAYOUT)

    WWDG.WWDG_CFR.WDGTB = 0b10
    WWDG.WWDG_CR.WDGA = 1
    print("Current counter:", WWDG.WWDG_CR.T)

Defining structure layout
-------------------------

Structure layout is defined by a "descriptor" - a Python dictionary which
encodes field names as keys and other properties required to access them as
associated values::

    {
        "field1": <properties>,
        "field2": <properties>,
        ...
    }

Currently, ``uctypes`` requires explicit specification of offsets for each
field. Offset are given in bytes from the structure start.

Following are encoding examples for various field types:

* Scalar types::

    "field_name": offset | uctypes.UINT32

  in other words, the value is a scalar type identifier ORed with a field offset
  (in bytes) from the start of the structure.

* Recursive structures::

    "sub": (offset, {
        "b0": 0 | uctypes.UINT8,
        "b1": 1 | uctypes.UINT8,
    })

  i.e. value is a 2-tuple, first element of which is an offset, and second is
  a structure descriptor dictionary (note: offsets in recursive descriptors
  are relative to the structure it defines). Of course, recursive structures
  can be specified not just by a literal dictionary, but by referring to a
  structure descriptor dictionary (defined earlier) by name.

* Arrays of primitive types::

      "arr": (offset | uctypes.ARRAY, size | uctypes.UINT8),

  i.e. value is a 2-tuple, first element of which is ARRAY flag ORed
  with offset, and second is scalar element type ORed number of elements
  in the array.

* Arrays of aggregate types::

    "arr2": (offset | uctypes.ARRAY, size, {"b": 0 | uctypes.UINT8}),

  i.e. value is a 3-tuple, first element of which is ARRAY flag ORed
  with offset, second is a number of elements in the array, and third is
  a descriptor of element type.

* Pointer to a primitive type::

    "ptr": (offset | uctypes.PTR, uctypes.UINT8),

  i.e. value is a 2-tuple, first element of which is PTR flag ORed
  with offset, and second is a scalar element type.

* Pointer to an aggregate type::

    "ptr2": (offset | uctypes.PTR, {"b": 0 | uctypes.UINT8}),

  i.e. value is a 2-tuple, first element of which is PTR flag ORed
  with offset, second is a descriptor of type pointed to.

* Bitfields::

    "bitf0": offset | uctypes.BFUINT16 | lsbit << uctypes.BF_POS | bitsize << uctypes.BF_LEN,

  i.e. value is a type of scalar value containing given bitfield (typenames are
  similar to scalar types, but prefixes with ``BF``), ORed with offset for
  scalar value containing the bitfield, and further ORed with values for
  bit position and bit length of the bitfield within the scalar value, shifted by
  BF_POS and BF_LEN bits, respectively. A bitfield position is counted
  from the least significant bit of the scalar (having position of 0), and
  is the number of right-most bit of a field (in other words, it's a number
  of bits a scalar needs to be shifted right to extract the bitfield).

  In the example above, first a UINT16 value will be extracted at offset 0
  (this detail may be important when accessing hardware registers, where
  particular access size and alignment are required), and then bitfield
  whose rightmost bit is *lsbit* bit of this UINT16, and length
  is *bitsize* bits, will be extracted. For example, if *lsbit* is 0 and
  *bitsize* is 8, then effectively it will access least-significant byte
  of UINT16.

  Note that bitfield operations are independent of target byte endianness,
  in particular, example above will access least-significant byte of UINT16
  in both little- and big-endian structures. But it depends on the least
  significant bit being numbered 0. Some targets may use different
  numbering in their native ABI, but ``uctypes`` always uses the normalized
  numbering described above.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final

from uio import AnyReadableBuf

_ScalarProperty: Final = int
_RecursiveProperty: Final = tuple[int, "_property"]
_ArrayProperty: Final = tuple[int, int]
_ArrayOfAggregateProperty: Final = tuple[int, int, "_property"]
_PointerToAPrimitiveProperty: Final = tuple[int, int]
_PointerToAaAggregateProperty: Final = tuple[int, "_property"]
_BitfieldProperty: Final = int
_property: Final = _ScalarProperty | _RecursiveProperty | _ArrayProperty | _ArrayOfAggregateProperty | _PointerToAPrimitiveProperty | _PointerToAaAggregateProperty | _BitfieldProperty
_descriptor: Final = tuple[str, _property]

LITTLE_ENDIAN: Final[int] = ...
"""
Layout type for a little-endian packed structure. (Packed means that every
   field occupies exactly as many bytes as defined in the descriptor, i.e.
   the alignment is 1).
"""

BIG_ENDIAN: Final[int] = ...
"""
Layout type for a big-endian packed structure.
"""

NATIVE: Final[int] = ...
"""
Layout type for a native structure - with data endianness and alignment
   conforming to the ABI of the system on which MicroPython runs.
"""

# noinspection PyShadowingNames
def sizeof(struct: struct | _descriptor, layout_type: int = NATIVE, /) -> int:
    """
   Return size of data structure in bytes. The *struct* argument can be
   either a structure class or a specific instantiated structure object
   (or its aggregate field).
   """

def addressof(obj: AnyReadableBuf, /) -> int:
    """
   Return address of an object. Argument should be bytes, bytearray or
   other object supporting buffer protocol (and address of this buffer
   is what actually returned).
   """

def bytes_at(addr: int, size: int, /) -> bytes:
    """
   Capture memory at the given address and size as bytes object. As bytes
   object is immutable, memory is actually duplicated and copied into
   bytes object, so if memory contents change later, created object
   retains original value.
   """

def bytearray_at(addr: int, size: int, /) -> bytearray:
    """
   Capture memory at the given address and size as bytearray object.
   Unlike bytes_at() function above, memory is captured by reference,
   so it can be both written too, and you will access current value
   at the given memory address.
   """

UINT8: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

INT8: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

UINT16: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

INT16: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

UINT32: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

INT32: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

UINT64: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

INT64: Final[int] = ...
"""
Integer types for structure descriptors. Constants for 8, 16, 32,
   and 64 bit types are provided, both signed and unsigned.
"""

FLOAT32: Final[int] = ...
"""
Floating-point types for structure descriptors.
"""

FLOAT64: Final[int] = ...
"""
Floating-point types for structure descriptors.
"""

VOID: Final[int] = ...
"""
``VOID`` is an alias for ``UINT8``, and is provided to conveniently define
   C's void pointers: ``(uctypes.PTR, uctypes.VOID)``.
"""

PTR: Final[int] = ...
"""
Type constants for pointers and arrays. Note that there is no explicit
   constant for structures, it's implicit: an aggregate type without ``PTR``
   or ``ARRAY`` flags is a structure.
"""

ARRAY: Final[int] = ...
"""
Type constants for pointers and arrays. Note that there is no explicit
   constant for structures, it's implicit: an aggregate type without ``PTR``
   or ``ARRAY`` flags is a structure.
"""

# noinspection PyPep8Naming
class struct:
    """
   Module contents
   ---------------
   """

    def __init__(
        self, addr: int, descriptor: _descriptor, layout_type: int = NATIVE, /
    ):
        """
      Instantiate a "foreign data structure" object based on structure address in
      memory, descriptor (encoded as a dictionary), and layout type (see below).
      """



import tkinter as tk
import traceback
from logging import getLogger
from typing import List, Optional, Tuple

from turcar import get_workbench
from turcar.codeview import CodeViewText, SyntaxText, get_syntax_options_for_tag
from turcar.common import SignatureInfo, SignatureParameter
from turcar.editors import Editor
from turcar.misc_utils import running_on_mac_os
from turcar.shell import ShellText
from turcar.tktextext import TextFrame
from turcar.ui_utils import get_tk_version_info

all_boxes = []

a_box_is_appearing = False

logger = getLogger(__name__)


class EditorInfoBox(tk.Toplevel):
    def __init__(self):
        super().__init__(master=get_workbench())
        self._has_shown_on_screen: bool = False
        all_boxes.append(self)
        self._target_text_widget: Optional[SyntaxText] = None

        get_workbench().bind("<FocusOut>", self._workbench_focus_out, True)

        # If the box has received focus, then it may lose it by a messagebox
        # or mouse click on the main window
        self.bind("<FocusOut>", self._workbench_focus_out, True)

        get_workbench().bind("<Escape>", self.hide, True)
        self.bind("<Escape>", self.hide, True)
        get_workbench().bind_class("EditorCodeViewText", "<1>", self.hide, True)
        get_workbench().bind_class("ShellText", "<1>", self.hide, True)

        get_workbench().bind("SyntaxThemeChanged", self._update_theme, True)

    def _set_window_attributes(self):
        if running_on_mac_os():
            try:
                # Must be the first thing to do after creating window
                # https://wiki.tcl-lang.org/page/MacWindowStyle
                self.tk.call(
                    "::tk::unsupported::MacWindowStyle", "style", self._w, "help", "noActivates"
                )
                if get_tk_version_info() >= (8, 6, 10) and running_on_mac_os():
                    self.wm_overrideredirect(1)
            except tk.TclError:
                pass
        else:
            self.wm_overrideredirect(1)
        self.wm_transient(get_workbench())

        # From IDLE
        # TODO: self.update_idletasks()  # Need for tk8.6.8 on macOS: #40128.
        self.lift()

    def _update_theme(self, event=None):
        pass

    def _check_bind_for_keypress(self, text: tk.Text):
        tag_prefix = "pb_" + type(self).__name__.replace(".", "_")
        if getattr(text, tag_prefix, False):
            return False

        # Need to know about certain keypresses while the completer is active
        # Other plugins (eg. auto indenter) may eat up returns, therefore I need to
        # raise the priority of this binding
        tag = tag_prefix + "_" + str(text.winfo_id())
        text.bindtags((tag,) + text.bindtags())
        text.bind_class(tag, "<Key>", self._on_text_keypress, True)

        setattr(text, tag_prefix, True)

    def _on_text_keypress(self, event=None):
        pass

    def _workbench_focus_out(self, event=None) -> None:
        if not self.is_visible():
            return

        # if a_box_is_appearing:
        # making a box appear may mess up FocusOut events
        #    return

        # Need to close when another app or a turcar's dialog appears
        # (othewise the box will float above this, at least in Linux).
        # Don't do anything if another EditorInfoBox appears
        for box in all_boxes:
            try:
                # need to try because asking for focus via wrong window may give exception
                if box.focus_get():
                    # it's alright
                    return
            except:
                pass

        self.hide()

    def _get_related_box(self) -> Optional["EditorInfoBox"]:
        return None

    def _show_on_target_text(
        self,
        index: str,
        expected_box_height: int,
        preferred_position: str,
        y_offset: int = 0,
    ) -> None:
        text = self._target_text_widget

        bbox = text.bbox(index)
        if not bbox:
            logger.warning("Could not compute bbox")
            return

        text_box_x, text_box_y, _, text_box_height = bbox

        cursor_root_x = text.winfo_rootx() + text_box_x
        cursor_root_y = text.winfo_rooty() + text_box_y

        if (
            preferred_position == "below"
            and cursor_root_y + text_box_height + expected_box_height > text.winfo_screenheight()
        ):
            position = "above"
        else:
            position = preferred_position

        if position == "above":
            # negative value signifies pixels between window bottom and screen bottom
            y = cursor_root_y - text.winfo_screenheight()
        else:
            y = cursor_root_y + text_box_height

        # TODO reduce x if the box wouldn't fit by width
        x = cursor_root_x
        self._show_on_screen(x, y + y_offset)

    def _show_on_screen(self, x: int, y: int) -> None:
        global a_box_is_appearing
        if a_box_is_appearing:
            logger.debug("Box already appearing, skipping _show_on_screen")
            return

        try:
            a_box_is_appearing = True

            if y < 0:
                self.geometry("+%d-%d" % (x, -y))
            else:
                self.geometry("+%d+%d" % (x, y))

            if not self.winfo_ismapped():
                self._set_window_attributes()
                self._check_update_size()
                self.deiconify()
                if not self._has_shown_on_screen:
                    self.tweak_first_appearance()
            else:
                self._check_update_size()
        finally:
            a_box_is_appearing = False

        self._has_shown_on_screen = True

    def _check_update_size(self) -> None:
        if hasattr(self, "_update_size"):
            # It looks it's not worth trying to move the window away from the viewport
            # for calculations. At least in Ubuntu it doesn't give any advantages and
            # may produce glitches
            # self.geometry("+10000+5000")  # move away from visible screen
            # self.withdraw()

            self.update()  # gives proper data for size calculations
            self._update_size()
            self.update()  # applies updated size

    def hide(self, event: Optional[tk.Event] = None) -> None:
        if self.winfo_ismapped():
            self.withdraw()

        related_box = self._get_related_box()
        if related_box and related_box.is_visible():
            related_box.hide(event)

        # Following looks like a good idea, but in at least in Ubuntu, it would fix
        # entry cursor to the given text and it can't be moved to another text anymore
        # if self._target_text_widget:
        #    self._target_text_widget.focus_set()  # in case the user has clicked on the box

    def is_visible(self) -> bool:
        return self.winfo_ismapped()

    def tweak_first_appearance(self):
        pass


class DocuBoxBase(EditorInfoBox):
    def __init__(self, show_vertical_scrollbar: bool):
        super().__init__()
        self.text_frame = TextFrame(
            master=self,
            horizontal_scrollbar=False,
            vertical_scrollbar=show_vertical_scrollbar,
            read_only=True,
            height=7,
            width=40,
            font="TkDefaultFont",
            wrap="word",
        )
        self.text_frame.grid()
        self.text = self.text_frame.text

        self._update_theme()

    def _update_theme(self, event=None):
        super()._update_theme(event)
        comment_opts = get_syntax_options_for_tag("comment")
        gutter_opts = get_syntax_options_for_tag("GUTTER")
        text_opts = get_syntax_options_for_tag("TEXT")
        self.text["background"] = gutter_opts["background"]
        self.text["foreground"] = text_opts["foreground"]
        self.text.tag_configure("prose", font="TkDefaultFont")
        self.text.tag_configure("active", font="BoldTkDefaultFont")
        self.text.tag_configure("annotation", **comment_opts)
        self.text.tag_configure("default", **comment_opts)
        self.text.tag_configure("marker", **comment_opts)

    def _append_chars(self, chars, tags=()):
        self.text.direct_insert("end", chars, tags=tuple(tags))

    def render_signatures(self, signatures: List[SignatureInfo], only_params=False) -> None:
        for i, sig in enumerate(signatures):
            if i > 0:
                self._append_chars("\n")
            self.render_signature(sig, only_params)

    def render_signature(self, sig: SignatureInfo, only_params) -> None:
        if not only_params:
            self._append_chars(sig.name)

        self._append_chars("(")

        is_positional = False
        is_kw_only = False

        for i, param in enumerate(sig.params):
            if i > 0:
                self._append_chars(", ")
            if len(sig.params) > 20:
                self._append_chars("\n    ")

            is_positional |= param.kind == "POSITIONAL_ONLY"
            if is_positional and param.kind != "POSITIONAL_ONLY":
                self._append_chars("/, ", ["marker"])
                is_positional = False

            if param.kind == "VAR_POSITIONAL":
                is_kw_only = True
            elif param.kind == "KEYWORD_ONLY" and not is_kw_only:
                self._append_chars("*, ", ["marker"])
                is_kw_only = True

            is_active_parameter = sig.current_param_index == i
            self.render_parameter(param, is_active_parameter)

        if is_positional:
            self._append_chars(", /", ["marker"])

        self._append_chars(")")

        if sig.return_type and not only_params:
            self._append_chars(" -> ", ["marker"])
            self._append_chars(sig.return_type, ["annotation"])

    def render_parameter(self, param: SignatureParameter, active: bool) -> None:
        if active:
            base_tags = ["active"]
        else:
            base_tags = []

        if param.kind == "VAR_POSITIONAL":
            self._append_chars("*", base_tags)
        elif param.kind == "VAR_KEYWORD":
            self._append_chars("**", base_tags)

        self._append_chars(param.name, base_tags)

        if param.annotation:
            self._append_chars(":\u00A0" + param.annotation, base_tags + ["annotation"])

        if param.default:
            self._append_chars("=" + param.default, base_tags + ["default"])

    def format_signature(self, s: str) -> str:
        s = s.replace(": ", ":\u00A0")
        if len(s) > self.text["width"] * 1.8 and s.count("(") and s.count(")"):
            args_index = s.index("(") + 1
            suffix_index = s.rindex(")")
            prefix = s[:args_index]
            args = s[args_index:suffix_index].split(", ")
            suffix = s[suffix_index:]
            s = prefix + "\n  " + ",\n  ".join(args) + "\n" + suffix
            # don't keep / and * alone on a line
            s = (
                s.replace("\n  /,", " /,")
                .replace("\n  *,", " *,")
                .replace("\n  /\n)", " /\n)")
                .replace("\n  *\n)", " *\n)")
            )
            return s
        else:
            return s


class DocuBox(DocuBoxBase):
    def __init__(self):
        super().__init__(show_vertical_scrollbar=True)

    def set_content(self, name, item_type, signatures, docstring):
        self.text.direct_delete("1.0", "end")

        # self._append_chars(item_type + "\n")

        if signatures:
            self.render_signatures(signatures)

        if signatures and docstring:
            self._append_chars("\n\n")

        if docstring:
            self._append_chars(docstring, ["prose"])


def get_active_text_widget() -> Optional[SyntaxText]:
    widget = get_workbench().focus_get()
    if isinstance(widget, (CodeViewText, ShellText)):
        return widget

    return None


def get_cursor_position(text: SyntaxText) -> Tuple[int, int]:
    parts = text.index("insert").split(".")
    return int(parts[0]), int(parts[1])


def get_text_filename(text: SyntaxText) -> Optional[str]:
    if isinstance(text, ShellText):
        return "<Shell>"
    elif isinstance(text, CodeViewText):
        editor = text.master.master
        if isinstance(editor, Editor):
            return editor.get_filename()

    return None


def get_relevant_source_and_cursor_position(text: SyntaxText) -> Tuple[str, int, int]:
    if isinstance(text, ShellText):
        source = text.get("input_start", "insert")
        lines = source.splitlines()
        if not lines:
            return source, 1, 0
        else:
            return source, len(lines), len(lines[-1])
    else:
        row, col = get_cursor_position(text)
        return text.get("1.0", "end-1c"), row, col



"""Support for camera input

The `camera` module contains classes to control the camera and take pictures."""

from __future__ import annotations

from circuitpython_typing import WriteableBuffer

class Camera:
    """The class to control camera.

    Usage::

        import board
        import sdioio
        import storage
        import camera

        sd = sdioio.SDCard(
            clock=board.SDIO_CLOCK,
            command=board.SDIO_COMMAND,
            data=board.SDIO_DATA,
            frequency=25000000)
        vfs = storage.VfsFat(sd)
        storage.mount(vfs, '/sd')

        cam = camera.Camera()

        buffer = bytearray(512 * 1024)
        file = open("/sd/image.jpg","wb")
        size = cam.take_picture(buffer, width=1920, height=1080, format=camera.ImageFormat.JPG)
        file.write(buffer, size)
        file.close()"""

    def __init__(self) -> None:
        """Initialize camera."""
        ...
    def deinit(self) -> None:
        """De-initialize camera."""
        ...
    def take_picture(self, buf: WriteableBuffer, format: ImageFormat) -> int:
        """Take picture and save to ``buf`` in the given ``format``. The size of the picture
        taken is ``width`` by ``height`` in pixels.

        :return: the number of bytes written into buf
        :rtype: int"""
        ...

class ImageFormat:
    """Image format"""

    def __init__(self) -> None:
        """Enum-like class to define the image format."""
    JPG: ImageFormat
    """JPG format."""

    RGB565: ImageFormat
    """RGB565 format."""



from typing import Sequence, TypeVar, overload

_T = TypeVar('_T')


def choice(seq: Sequence[_T]) -> _T:
    """
    Return a random element from the non-empty sequence seq. If seq is empty, raises IndexError.
    """


def getrandbits(n: int) -> int:
    """
    Returns an integer with n random bits.

    Because the underlying generator function returns at most 30 bits, n may only be a value between 1-30 (inclusive).
    """


def randint(a: int, b: int) -> int:
    """
    Return a random integer N such that a <= N <= b. Alias for randrange(a, b+1).
    """


def random() -> float:
    """
    Return the next random floating point number in the range [0.0, 1.0).
    """


@overload
def randrange(stop: int) -> int:
    """
    Return a randomly selected integer between zero and up to (but not including) stop.
    """

@overload
def randrange(start: int, stop: int) -> int:
    """
    Return a randomly selected integer from range(start, stop).
    """

@overload
def randrange(start: int, stop: int, step: int) -> int:
    """
    Return a randomly selected element from range(start, stop, step).
    """


def seed(n: int) -> None:
    """
    Initialize the random number generator with a known integer n. This will give you reproducibly deterministic randomness from a given starting state (n).
    """


def uniform(a: float, b: float) -> float:
    """
    Return a random floating point number N such that a <= N <= b for a <= b and b <= N <= a for b < a.
    """



from enum import Enum


class ImageFile(Enum):
    """
    Paths to standard EV3 images.

    Note:
        Information: ACCEPT, BACKWARD, DECLINE, FORWARD, LEFT, NO_GO, QUESTION_MARK, RIGHT, STOP_1, STOP_2, THUMBS_DOWN, THUMBS_UP, WARNING

        LEGO&reg;: EV3, EV3_ICON

        Objects: TARGET

        Eyes: ANGRY, AWAKE, BOTTOM_LEFT, BOTTOM_RIGHT, CRAZY_1, CRAZY_2, DIZZY, DOWN, EVIL, KNOCKED_OUT, MIDDLE_LEFT, MIDDLE_RIGHT, NEUTRAL, PINCHED_LEFT, PINCHED_MIDDLE, PINCHED_RIGHT, SLEEPING, TIRED_LEFT, TIRED_MIDDLE, TIRED_RIGHT, UP, WINKING
    """
    ACCEPT = "ACCEPT"
    BACKWARD = "BACKWARD"
    DECLINE = "DECLINE"
    FORWARD = "FORWARD"
    LEFT = "LEFT"
    NO_GO = "NO_GO"
    QUESTION_MARK = "QUESTION_MARK"
    RIGHT = "RIGHT"
    STOP_1 = "STOP_1"
    STOP_2 = "STOP_2"
    THUMBS_DOWN = "THUMBS_DOWN"
    THUMBS_UP = "THUMBS_UP"
    WARNING = "WARNING"

    EV3 = "EV3"
    EV3_ICON = "EV3_ICON"

    TARGET = "TARGET"

    ANGRY = "ANGRY"
    AWAKE = "AWAKE"
    BOTTOM_LEFT = "BOTTOM_LEFT"
    BOTTOM_RIGHT = "BOTTOM_RIGHT"
    CRAZY_1 = "CRAZY_1"
    CRAZY_2 = "CRAZY_2"
    DIZZY = "DIZZY"
    DOWN = "DOWN"
    EVIL = "EVIL"
    KNOCKED_OUT = "KNOCKED_OUT"
    MIDDLE_LEFT = "MIDDLE_LEFT"
    MIDDLE_RIGHT = "MIDDLE_RIGHT"
    NEUTRAL = "NEUTRAL"
    PINCHED_LEFT = "PINCHED_LEFT"
    PINCHED_MIDDLE = "PINCHED_MIDDLE"
    PINCHED_RIGHT = "PINCHED_RIGHT"
    SLEEPING = "SLEEPING"
    TIRED_LEFT = "TIRED_LEFT"
    TIRED_MIDDLE = "TIRED_MIDDLE"
    TIRED_RIGHT = "TIRED_RIGHT"
    WINKING = "WINKING"
    UP = "UP"



def disable_irq():
    pass


def enable_irq():
    pass


def freq():
    pass


mem16 = None
mem32 = None
mem8 = None


def reset():
    pass


def time_pulse_us():
    pass


def unique_id():
    pass



"""AES encryption routines

The `AES` module contains classes used to implement encryption
and decryption. It aims to be low overhead in terms of memory.

For more information on AES, refer to `the Wikipedia entry
<https://en.wikipedia.org/wiki/Advanced_Encryption_Standard>`_.
"""

from __future__ import annotations

from typing import Optional

from circuitpython_typing import ReadableBuffer, WriteableBuffer

MODE_ECB: int
MODE_CBC: int
MODE_CTR: int

class AES:
    """Encrypt and decrypt AES streams"""

    def __init__(
        self,
        key: ReadableBuffer,
        mode: int = 0,
        IV: Optional[ReadableBuffer] = None,
        segment_size: int = 8,
    ) -> None:
        """Create a new AES state with the given key.

        :param ~circuitpython_typing.ReadableBuffer key: A 16-, 24-, or 32-byte key
        :param int mode: AES mode to use.  One of: `MODE_ECB`, `MODE_CBC`, or
                         `MODE_CTR`
        :param ~circuitpython_typing.ReadableBuffer IV: Initialization vector to use for CBC or CTR mode

        Additional arguments are supported for legacy reasons.

        Encrypting a string::

          import aesio
          from binascii import hexlify

          key = b'Sixteen byte key'
          inp = b'CircuitPython!!!' # Note: 16-bytes long
          outp = bytearray(len(inp))
          cipher = aesio.AES(key, aesio.MODE_ECB)
          cipher.encrypt_into(inp, outp)
          hexlify(outp)"""
        ...
    def rekey(
        self,
        key: ReadableBuffer,
        IV: Optional[ReadableBuffer] = None,
    ) -> None:
        """Update the AES state with the given key.

        :param ~circuitpython_typing.ReadableBuffer key: A 16-, 24-, or 32-byte key
        :param ~circuitpython_typing.ReadableBuffer IV: Initialization vector to use
                                                        for CBC or CTR mode"""
        ...
    def encrypt_into(self, src: ReadableBuffer, dest: WriteableBuffer) -> None:
        """Encrypt the buffer from ``src`` into ``dest``.

        For ECB mode, the buffers must be 16 bytes long.  For CBC mode, the
        buffers must be a multiple of 16 bytes, and must be equal length.  For
        CTX mode, there are no restrictions."""
        ...
    def decrypt_into(self, src: ReadableBuffer, dest: WriteableBuffer) -> None:
        """Decrypt the buffer from ``src`` into ``dest``.
        For ECB mode, the buffers must be 16 bytes long.  For CBC mode, the
        buffers must be a multiple of 16 bytes, and must be equal length.  For
        CTX mode, there are no restrictions."""
        ...



class BluetoothMailboxClient:
    """
    Object that represents a Bluetooth connection to one or more remote EV3s.

    The remote EV3s can either be running MicroPython or the standard EV3 firmware.

    A “client” initiates a connection to a waiting “server”.
    """

    def __init__(self):
        ...

    def connect(self, brick: str):
        """
        Connects to a BluetoothMailboxServer on another device.

        The remote device must be paired and waiting for a connection. See BluetoothMailboxServer.wait_for_connection().

        Args:
            brick (str): The name or Bluetooth address of the remove EV3 to connect to.

        Raises:
            OSError: There was a problem establishing the connection.
        """
        ...

    def close(self):
        """
        Closes all connections.
        """
        ...



"""Low-level bit primitives for Maxim (formerly Dallas Semi) one-wire protocol.

   Protocol definition is here: https://www.maximintegrated.com/en/app-notes/index.mvp/id/126"""

from __future__ import annotations

import microcontroller

class OneWire:
    def __init__(self, pin: microcontroller.Pin) -> None:
        """Create a OneWire object associated with the given pin.

        The object implements the lowest level timing-sensitive bits of the protocol.

        :param ~microcontroller.Pin pin: Pin connected to the OneWire bus

        Read a short series of pulses::

          import onewireio
          import board

          onewire = onewireio.OneWire(board.D7)
          onewire.reset()
          onewire.write_bit(True)
          onewire.write_bit(False)
          print(onewire.read_bit())"""
        ...
    def deinit(self) -> None:
        """Deinitialize the OneWire bus and release any hardware resources for reuse."""
        ...
    def __enter__(self) -> OneWire:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def reset(self) -> bool:
        """Reset the OneWire bus and read presence

        :returns: False when at least one device is present
        :rtype: bool"""
        ...
    def read_bit(self) -> bool:
        """Read in a bit

        :returns: bit state read
        :rtype: bool"""
        ...
    def write_bit(self, value: bool) -> None:
        """Write out a bit based on value."""
        ...



from enum import Enum


class Color(Enum):
    """
    Light or surface color.
    """
    BLACK = ...
    BLUE = ...
    GREEN = ...
    YELLOW = ...
    RED = ...
    WHITE = ...
    BROWN = ...
    ORANGE = ...
    PURPLE = ...



import shlex

from .connection import MicroPythonConnection


class SshProcessConnection(MicroPythonConnection):
    def __init__(self, client, cwd, cmd):
        super().__init__()
        import threading

        self._client = client

        cmd_line_str = (
            "echo $$ ;"
            + ((" cd %s  2> /dev/null ;" % shlex.quote(cwd) if cwd else ""))
            + (" exec " + " ".join(map(shlex.quote, cmd)))
        )
        self._stdin, self._stdout, _ = self._client.exec_command(
            cmd_line_str, bufsize=0, timeout=None, get_pty=True
        )

        # stderr gets directed to stdout because of pty
        self._pid = self._stdout.readline().strip()

        self._reading_thread = threading.Thread(target=self._listen_output, daemon=True)
        self._reading_thread.start()

    def write(self, data: bytes) -> int:
        if isinstance(data, str):
            data = data.encode(self.encoding)
        self._stdin.write(data)
        self._stdin.flush()
        return len(data)

    def _listen_output(self):
        "NB! works in background thread"
        try:
            while not self._reader_stopped:
                data = self._stdout.read(1)
                if len(data) > 0:
                    self._make_output_available(data)
                else:
                    self._error = "EOF"
                    break

        except Exception as e:
            self._error = str(e)

    def close(self):
        self._client.exec_command("kill -s SIGKILL %s" % self._pid)
        self._reading_thread.join()
        self._client = None
        self._reading_thread = None



from typing import Callable


class Mailbox:
    """
    Object that represents a mailbox containing data.

    You can read data that is delivered by other EV3 bricks, or send data to other bricks that have the same mailbox.

    By default, the mailbox reads and send only bytes. To send other data, you can provide an encode function that encodes your Python object into bytes, and a decode function to convert bytes back to a Python object.

    Args:
        name (str): The name of this mailbox.
        connection: A connection object such as BluetoothMailboxClient.
        encode (callable): Function that encodes a Python object to bytes.
        decode (callable): Function that creates a new Python object from bytes.
    """

    def __init__(self, name: str, connection: object, encode: Callable = None, decode: Callable = None):
        ...

    def read(self) -> any:
        """
        Gets the current value of the mailbox.

        Returns:
            The current value or None if the mailbox is empty.
        """
        return ''

    def send(self, value: any, brick: str):
        """
        Sends a value to this mailbox on connected devices.

        Args:
            value (any): The value that will be delivered to the mailbox.
            brick (str): The name or Bluetooth address of the brick or None to broadcast to all connected devices.

        Raises:
            OSError: There is a problem with the connection.
        """
        ...

    def wait(self):
        """
        Waits for the mailbox to be updated by remote device.
        """
        ...

    def wait_new(self) -> any:
        """
        Waits for a new value to be delivered to the mailbox that is not equal to the current value in the mailbox.

        Returns:
            The new value.
        """
        return ''



from pybricks.parameters import Port


class LUMPDevice:
    """
    Devices using the LEGO UART Messaging Protocol.

    Args:
        port (Port): Port to which the device is connected.
    """

    def __init__(self, port: Port):
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def read(self, mode: int) -> tuple:
        """
        Reads values from a given mode.

        Args:
            mode (int): Device mode.

        Returns:
            Values read from the sensor.
        """
        return ()

    def write(self, mode: int, data: tuple):
        """
        Writes values to the sensor. Only selected sensors and modes support this.

        Args:
            mode (int): Device mode.
            data (tuple): Values to be written.
        """
        ...



"""
pack and unpack primitive data types

Descriptions taken from 
`https://raw.githubusercontent.com/micropython/micropython/master/docs/library/struct.rst`, etc.
=====================================================

.. module:: struct
   :synopsis: pack and unpack primitive data types

|see_cpython_module| :mod:`python:struct`.

Supported size/byte order prefixes: ``@``, ``<``, ``>``, ``!``.

Supported format codes: ``b``, ``B``, ``h``, ``H``, ``i``, ``I``, ``l``,
``L``, ``q``, ``Q``, ``s``, ``P``, ``f``, ``d`` (the latter 2 depending
on the floating-point support).

.. admonition:: Difference to CPython
   :class: attention

   Whitespace is not supported in format strings.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.1.0"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Any

from uio import AnyReadableBuf, AnyWritableBuf

def calcsize(fmt: str | bytes, /,) -> int:
    """
   Return the number of bytes needed to store the given *fmt*.
   """

def pack(fmt: str | bytes, /, *v: Any) -> bytes:
    """
   Pack the values *v1*, *v2*, ... according to the format string *fmt*.
   The return value is a bytes object encoding the values.
   """

def pack_into(
    fmt: str | bytes, buffer: AnyWritableBuf, offset: int, /, *v: Any
) -> None:
    """
   Pack the values *v1*, *v2*, ... according to the format string *fmt*
   into a *buffer* starting at *offset*. *offset* may be negative to count
   from the end of *buffer*.
   """

def unpack(fmt: str | bytes, data: AnyReadableBuf, /) -> tuple[Any, ...]:
    """
   Unpack from the *data* according to the format string *fmt*.
   The return value is a tuple of the unpacked values.
   """

def unpack_from(
    fmt: str | bytes, data: AnyReadableBuf, offset: int = 0, /
) -> tuple[Any, ...]:
    """
   Unpack from the *data* starting at *offset* according to the format string
   *fmt*. *offset* may be negative to count from the end of *buffer*. The return
   value is a tuple of the unpacked values.
   """



"""Alarms and sleep

Provides alarms that trigger based on time intervals or on external events, such as pin
changes.
The program can simply wait for these alarms, or go to sleep and be awoken when they trigger.

There are two supported levels of sleep: light sleep and deep sleep.

Light sleep keeps sufficient state so the program can resume after sleeping.
It does not shut down WiFi, BLE, or other communications, or ongoing activities such
as audio playback. It reduces power consumption to the extent possible that leaves
these continuing activities running. In some cases there may be no decrease in power consumption.

Deep sleep shuts down power to nearly all of the microcontroller including the CPU and RAM. This can save
a more significant amount of power, but CircuitPython must restart ``code.py`` from the beginning when
awakened.

For both light sleep and deep sleep, if CircuitPython is connected to a host computer,
maintaining the connection takes priority and power consumption may not be reduced.

For more information about working with alarms and light/deep sleep in CircuitPython,
see `this Learn guide <https://learn.adafruit.com/deep-sleep-with-circuitpython>`_.
"""

from __future__ import annotations

from typing import Optional, Sequence, overload

import circuitpython_typing
import digitalio
from circuitpython_typing import ReadableBuffer

sleep_memory: SleepMemory
"""Memory that persists during deep sleep.
This object is the sole instance of `alarm.SleepMemory`."""
wake_alarm: Optional[circuitpython_typing.Alarm]
"""The most recently triggered alarm. If CircuitPython was sleeping, the alarm that woke it from sleep.
If no alarm occurred since the last hard reset or soft restart, value is ``None``.
"""

def light_sleep_until_alarms(
    *alarms: circuitpython_typing.Alarm,
) -> circuitpython_typing.Alarm:
    """Go into a light sleep until awakened one of the alarms. The alarm causing the wake-up
    is returned, and is also available as `alarm.wake_alarm`.

    If no alarms are specified, return immediately.

    **If CircuitPython is connected to a host computer, the connection will be maintained,
    and the microcontroller may not actually go into a light sleep.**
    This allows the user to interrupt an existing program with ctrl-C,
    and to edit the files in CIRCUITPY, which would not be possible in true light sleep.
    Thus, to use light sleep and save significant power,
    it may be necessary to disconnect from the host.
    """
    ...

def exit_and_deep_sleep_until_alarms(
    *alarms: circuitpython_typing.Alarm,
    preserve_dios: Sequence[digitalio.DigitalInOut] = (),
) -> None:
    """Exit the program and go into a deep sleep, until awakened by one of the alarms.
    This function does not return.

    When awakened, the microcontroller will restart and will run ``boot.py`` and ``code.py``
    from the beginning.

    After restart, an alarm *equivalent* to the one that caused the wake-up
    will be available as `alarm.wake_alarm`.
    Its type and/or attributes may not correspond exactly to the original alarm.
    For time-base alarms, currently, an `alarm.time.TimeAlarm()` is created.

    If no alarms are specified, the microcontroller will deep sleep until reset.

    :param circuitpython_typing.Alarm alarms: the alarms that can wake the microcontroller.
    :param Sequence[digitalio.DigitalInOut] preserve_dios: A sequence of `DigitalInOut` objects
      whose state should be preserved during deep sleep.
      If a `DigitalInOut` in the sequence is set to be an output,
      its current `DigitalInOut.value` (``True`` or ``False``)
      will be preserved during the deep sleep.
      If a `DigitalInOut` in the sequence is set to be an input,
      its current `DigitalInOut.pull` value (``DOWN``, ``UP``, or ``None``)
      will be preserved during deep sleep.

    Preserving `DigitalInOut` states during deep sleep can be used to ensure that
    external or on-board devices are powered or unpowered during sleep, among other purposes.

    On some microcontrollers, some pins cannot remain in their original state for hardware reasons.

    **Limitations:** ``preserve_dios`` is currently only available on Espressif.

    .. note::
      On Espressif chips, preserving pin settings during deep sleep may consume extra current.
      On ESP32, this was measured to be 250 uA or more.
      Consider not preserving pins unless you need to.
      Measure power consumption carefully both with no pins preserved and with the pins you might want to
      preserve to achieve the lowest consumption.

    **If CircuitPython is connected to a host computer via USB or BLE
    the first time a deep sleep is requested,
    the connection will be maintained and the system will not go into deep sleep.**
    This allows the user to interrupt an existing program with ctrl-C,
    and to edit the files in CIRCUITPY, which would not be possible in true deep sleep.

    If CircuitPython goes into a true deep sleep, and USB or BLE is reconnected,
    the next deep sleep will still be a true deep sleep. You must do a hard reset
    or power-cycle to exit a true deep sleep loop.

    Here is a skeletal example:

    .. code-block:: python

        import alarm
        import time
        import board

        print("Waking up")

        # Create an alarm for 60 seconds from now, and also a pin alarm.
        time_alarm = alarm.time.TimeAlarm(monotonic_time=time.monotonic() + 60)
        pin_alarm = alarm.pin.PinAlarm(board.D7, False)

        # Deep sleep until one of the alarm goes off. Then restart the program.
        alarm.exit_and_deep_sleep_until_alarms(time_alarm, pin_alarm)
    """
    ...

class SleepMemory:
    """Store raw bytes in RAM that persists during deep sleep.
    The class acts as a ``bytearray``.
    If power is lost, the memory contents are lost.

    Note that this class can't be imported and used directly. The sole
    instance of :class:`SleepMemory` is available at
    :attr:`alarm.sleep_memory`.

    **Limitations:** Not supported on RP2040.

    Usage::

       import alarm
       alarm.sleep_memory[0] = True
       alarm.sleep_memory[1] = 12
    """

    def __init__(self) -> None:
        """Not used. Access the sole instance through `alarm.sleep_memory`."""
        ...
    def __bool__(self) -> bool:
        """``sleep_memory`` is ``True`` if its length is greater than zero.
        This is an easy way to check for its existence.
        """
        ...
    def __len__(self) -> int:
        """Return the length. This is used by (`len`)"""
        ...
    @overload
    def __getitem__(self, index: slice) -> bytearray: ...
    @overload
    def __getitem__(self, index: int) -> int:
        """Returns the value at the given index."""
        ...
    @overload
    def __setitem__(self, index: slice, value: ReadableBuffer) -> None: ...
    @overload
    def __setitem__(self, index: int, value: int) -> None:
        """Set the value at the given index."""
        ...



from pybricks.media.ev3dev import SoundFile
from typing import Iterable, Union


class Speaker:
    """
    A stub class to represent the speaker member of the EV3Brick class.
    """

    def __init__(self):
        ...

    def beep(self, frequency: int = 500, duration: int = 100):
        """
        Play a beep/tone.

        Args:
            frequency (int): Frequency of the beep in Hertz. Frequencies below 100 are treated as 100.
            duration (int): Duration of the beep in milliseconds. If the duration is less than 0, then the method returns immediately, and the frequency play continues to play indefinitely.
        """
        ...

    def play_notes(self, notes: Iterable[str], tempo: int = 120):
        """
        Plays a sequence of musical notes.

        For example, you can play: ['C4/4', 'C4/4', 'G4/4', 'G4/4'].

        Args:
            notes (Iterable[str]): A sequence of notes to be played (see format below).
            tempo (int): Beats per minute where a quarter note is one beat.

        Note:
            Each note is a string with the following format:
                - The first character is the name of the note, A to G or R for a rest.
                # (sharp) or b (flat). B#/Cb and E#/Fb are not allowed.
                - Note names can also include an accidental
                - The note name is followed by the octave number 2 to 8. For example C4 is middle C. The octave changes to the next number at the note C, for example, B3 is the note below middle C (C4).
                - The octave is followed by / and a number that indicates the size of the note. For example /4 is a quarter note, /8 is an eighth note and so on.
                - This can optionally followed by a . to make a dotted note. Dotted notes are 1-1/2 times as long as notes without a dot.
                - The note can optionally end with a _ which is a tie or a slur. This causes there to be no pause between this note and the next note.
        """
        ...

    def play_file(self, file_name: Union[str, SoundFile]):
        """
        Plays a sound file.

        Args:
            file_name (str or SoundFile): Path to the sound file, including the file extension.
        """
        ...

    def say(self, text: str):
        """
        Says a given text string.

        You can configure the language and voice of the text using set_speech_options().

        Args:
            text (str): What to say.
        """
        ...

    def set_speech_options(self, language: str = None, voice: str = None, speed: int = None, pitch: int = None):
        """
        Configures speech settings used by the say() method.

        Any options that is set to None will not be changed. If an option is set to an invalid value say() will use the default value instead.
        
        Args:
            language (str): Language of the text. For example, you can choose 'en' (English) or 'de' (German). A list of all available languages is given below.
            voice (str): The voice to use. For example, you can choose 'f1' (female voice variant 1) or 'm3' (male voice variant 3). A list of all available voices is given below.
            speed (int): Number of words per minute.
            pitch (int): Pitch (0 to 99). Higher numbers make the voice higher pitched and lower numbers make the voice lower pitched.

        Note:
            You can choose the following languages:
                - 'af': Afrikaans
                - 'an': Aragonese
                - 'bg': Bulgarian
                - 'bs': Bosnian
                - 'ca': Catalan
                - 'cs': Czech
                - 'cy': Welsh
                - 'da': Danish
                - 'de': German
                - 'el': Greek
                - 'en': English (default)
                - 'en-gb': English (United Kingdom)
                - 'en-sc': English (Scotland)
                - 'en-uk-north': English (United Kingdom, Northern)
                - 'en-uk-rp': English (United Kingdom, Received Pronunciation)
                - 'en-uk-wmids': English (United Kingdom, West Midlands)
                - 'en-us': English (United States)
                - 'en-wi': English (West Indies)
                - 'eo': Esperanto
                - 'es': Spanish
                - 'es-la': Spanish (Latin America)
                - 'et': Estonian
                - 'fa': Persian
                - 'fa-pin': Persian
                - 'fi': Finnish
                - 'fr-be': French (Belgium)
                - 'fr-fr': French (France)
                - 'ga': Irish
                - 'grc': Greek
                - 'hi': Hindi
                - 'hr': Croatian
                - 'hu': Hungarian
                - 'hy': Armenian
                - 'hy-west': Armenian (Western)
                - 'id': Indonesian
                - 'is': Icelandic
                - 'it': Italian
                - 'jbo': Lojban
                - 'ka': Georgian
                - 'kn': Kannada
                - 'ku': Kurdish
                - 'la': Latin
                - 'lfn': Lingua Franca Nova
                - 'lt': Lithuanian
                - 'lv': Latvian
                - 'mk': Macedonian
                - 'ml': Malayalam
                - 'ms': Malay
                - 'ne': Nepali
                - 'nl': Dutch
                - 'no': Norwegian
                - 'pa': Punjabi
                - 'pl': Polish
                - 'pt-br': Portuguese (Brazil)
                - 'pt-pt': Portuguese (Portugal)
                - 'ro': Romanian
                - 'ru': Russian
                - 'sk': Slovak
                - 'sq': Albanian
                - 'sr': Serbian
                - 'sv': Swedish
                - 'sw': Swahili
                - 'ta': Tamil
                - 'tr': Turkish
                - 'vi': Vietnamese
                - 'vi-hue': Vietnamese (Hue)
                - 'vi-sgn': Vietnamese (Saigon)
                - 'zh': Mandarin Chinese
                - 'zh-yue': Cantonese Chinese

            You can choose the following voices:
                - 'f1': female variant 1
                - 'f2': female variant 2
                - 'f3': female variant 3
                - 'f4': female variant 4
                - 'f5': female variant 5
                - 'm1': male variant 1
                - 'm2': male variant 2
                - 'm3': male variant 3
                - 'm4': male variant 4
                - 'm5': male variant 5
                - 'm6': male variant 6
                - 'm7': male variant 7
                - 'croak': croak
                - 'whisper': whisper
                - 'whisperf': female whisper
        """
        ...

    def set_volume(self, volume: int, which: str = "_all_"):
        """
        Sets the speaker volume.

        Args:
            volume (int): Volume of the speaker as a percentage (0 to 100).
            which (str): Which volume to set. 'Beep' sets the volume for beep() and play_notes(). 'PCM' sets the volume for play_file() and say(). '_all_' sets both at the same time.
        """
        ...



"""
network configuration.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/network.rst.
****************************************

.. module:: network
   :synopsis: network configuration

This module provides network drivers and routing configuration. To use this
module, a MicroPython variant/build with network capabilities must be installed.
Network drivers for specific hardware are available within this module and are
used to configure hardware network interface(s). Network services provided
by configured interfaces are then available for use via the :mod:`socket`
module.

For example::

    # connect/ show IP config a specific network interface
    # see below for examples of specific drivers
    import network
    import time
    nic = network.Driver(...)
    if not nic.isconnected():
        nic.connect()
        print("Waiting for connection...")
        while not nic.isconnected():
            time.sleep(1)
    print(nic.ifconfig())

    # now use socket as usual
    import socket
    addr = socket.getaddrinfo('micropython.org', 80)[0][-1]
    s = socket.socket()
    s.connect(addr)
    s.send(b'GET / HTTP/1.1\r\nHost: micropython.org\r\n\r\n')
    data = s.recv(1000)
    s.close()
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from abc import abstractmethod
from typing import Protocol, Callable, overload, Any, ClassVar, Final

import pyb

MODE_11B: Final[int] = ...
"""IEEE 802.11b"""

MODE_11G: Final[int] = ...
"""IEEE 802.11g"""

MODE_11N: Final[int] = ...
"""IEEE 802.11n"""
@overload
def phy_mode(self) -> int:
    """
    Get or set the PHY mode.
    
    If the *mode* parameter is provided, sets the mode to its value. If
    the function is called without parameters, returns the current mode.
    
    The possible modes are defined as constants:
        * ``MODE_11B`` -- IEEE 802.11b,
        * ``MODE_11G`` -- IEEE 802.11g,
        * ``MODE_11N`` -- IEEE 802.11n.
    
    Availability: ESP8266.
   """

@overload
def phy_mode(self, mode: int, /) -> None:
    """
    Get or set the PHY mode.
    
    If the *mode* parameter is provided, sets the mode to its value. If
    the function is called without parameters, returns the current mode.
    
    The possible modes are defined as constants:
        * ``MODE_11B`` -- IEEE 802.11b,
        * ``MODE_11G`` -- IEEE 802.11g,
        * ``MODE_11N`` -- IEEE 802.11n.
    
    Availability: ESP8266.
   """

class AbstractNIC(Protocol):
    """
   Common network adapter interface
   ================================
   
   This section describes an (implied) abstract base class for all network
   interface classes implemented by :term:`MicroPython ports <MicroPython port>`
   for different hardware. This means that MicroPython does not actually
   provide ``AbstractNIC`` class, but any actual NIC class, as described
   in the following sections, implements methods as described here.
   """

    @abstractmethod
    def __init__(self, id: Any = None, /, *args: Any, **kwargs: Any):
        """
      Instantiate a network interface object. Parameters are network interface
      dependent. If there are more than one interface of the same type, the first
      parameter should be `id`.
      """
    @overload
    @abstractmethod
    def active(self, /) -> bool:
        """
           Activate ("up") or deactivate ("down") the network interface, if
           a boolean argument is passed. Otherwise, query current state if
           no argument is provided. Most other methods require an active
           interface (behaviour of calling them on inactive interface is
           undefined).
      """
    @overload
    @abstractmethod
    def active(self, is_active: bool, /) -> None:
        """
           Activate ("up") or deactivate ("down") the network interface, if
           a boolean argument is passed. Otherwise, query current state if
           no argument is provided. Most other methods require an active
           interface (behaviour of calling them on inactive interface is
           undefined).
      """
    @overload
    @abstractmethod
    def connect(self, key: str | None = None, /, **kwargs: Any) -> None:
        """
          Connect the interface to a network. This method is optional, and
          available only for interfaces which are not "always connected".
          If no parameters are given, connect to the default (or the only)
          service. If a single parameter is given, it is the primary identifier
          of a service to connect to. It may be accompanied by a key
          (password) required to access said service. There can be further
          arbitrary keyword-only parameters, depending on the networking medium
          type and/or particular device. Parameters can be used to: a)
          specify alternative service identifier types; b) provide additional
          connection parameters. For various medium types, there are different
          sets of predefined/recommended parameters, among them:
          
          * WiFi: *bssid* keyword to connect to a specific BSSID (MAC address)
      """
    @overload
    @abstractmethod
    def connect(
        self, service_id: Any, key: str | None = None, /, **kwargs: Any
    ) -> None:
        """
          Connect the interface to a network. This method is optional, and
          available only for interfaces which are not "always connected".
          If no parameters are given, connect to the default (or the only)
          service. If a single parameter is given, it is the primary identifier
          of a service to connect to. It may be accompanied by a key
          (password) required to access said service. There can be further
          arbitrary keyword-only parameters, depending on the networking medium
          type and/or particular device. Parameters can be used to: a)
          specify alternative service identifier types; b) provide additional
          connection parameters. For various medium types, there are different
          sets of predefined/recommended parameters, among them:
          
          * WiFi: *bssid* keyword to connect to a specific BSSID (MAC address)
      """
    @abstractmethod
    def disconnect(self) -> None:
        """
          Disconnect from network.
      """
    @abstractmethod
    def isconnected(self) -> bool:
        """
          Returns ``True`` if connected to network, otherwise returns ``False``.
      """
    @abstractmethod
    def scan(self, **kwargs: Any) -> list[tuple[str, ...]]:
        """
          Scan for the available network services/connections. Returns a
          list of tuples with discovered service parameters. For various
          network media, there are different variants of predefined/
          recommended tuple formats, among them:
          
          * WiFi: (ssid, bssid, channel, RSSI, authmode, hidden). There
            may be further fields, specific to a particular device.
          
          The function may accept additional keyword arguments to filter scan
          results (e.g. scan for a particular service, on a particular channel,
          for services of a particular set, etc.), and to affect scan
          duration and other parameters. Where possible, parameter names
          should match those in connect().
      """
    @overload
    @abstractmethod
    def status(self) -> Any:
        """
          Query dynamic status information of the interface.  When called with no
          argument the return value describes the network link status.  Otherwise
          *param* should be a string naming the particular status parameter to
          retrieve.
          
          The return types and values are dependent on the network
          medium/technology.  Some of the parameters that may be supported are:
          
          * WiFi STA: use ``'rssi'`` to retrieve the RSSI of the AP signal
          * WiFi AP: use ``'stations'`` to retrieve a list of all the STAs
            connected to the AP.  The list contains tuples of the form
            (MAC, RSSI).
      """
    @overload
    @abstractmethod
    def status(self, param: str, /) -> Any:
        """
          Query dynamic status information of the interface.  When called with no
          argument the return value describes the network link status.  Otherwise
          *param* should be a string naming the particular status parameter to
          retrieve.
          
          The return types and values are dependent on the network
          medium/technology.  Some of the parameters that may be supported are:
          
          * WiFi STA: use ``'rssi'`` to retrieve the RSSI of the AP signal
          * WiFi AP: use ``'stations'`` to retrieve a list of all the STAs
            connected to the AP.  The list contains tuples of the form
            (MAC, RSSI).
      """
    @overload
    @abstractmethod
    def ifconfig(self) -> tuple[str, str, str, str]:
        """
          Get/set IP-level network interface parameters: IP address, subnet mask,
          gateway and DNS server. When called with no arguments, this method returns
          a 4-tuple with the above information. To set the above values, pass a
          4-tuple with the required information.  For example::
          
           nic.ifconfig(('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    @abstractmethod
    def ifconfig(self, ip_mask_gateway_dns: tuple[str, str, str, str], /) -> None:
        """
          Get/set IP-level network interface parameters: IP address, subnet mask,
          gateway and DNS server. When called with no arguments, this method returns
          a 4-tuple with the above information. To set the above values, pass a
          4-tuple with the required information.  For example::
          
           nic.ifconfig(('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    @abstractmethod
    def config(self, param: str, /) -> Any:
        """
          Get or set general network interface parameters. These methods allow to work
          with additional parameters beyond standard IP configuration (as dealt with by
          `ifconfig()`). These include network-specific and hardware-specific
          parameters. For setting parameters, the keyword argument
          syntax should be used, and multiple parameters can be set at once. For
          querying, a parameter name should be quoted as a string, and only one
          parameter can be queried at a time::
          
           # Set WiFi access point name (formally known as ESSID) and WiFi channel
           ap.config(essid='My AP', channel=11)
           # Query params one by one
           print(ap.config('essid'))
           print(ap.config('channel'))
      """
    @overload
    @abstractmethod
    def config(self, **kwargs: Any) -> None:
        """
          Get or set general network interface parameters. These methods allow to work
          with additional parameters beyond standard IP configuration (as dealt with by
          `ifconfig()`). These include network-specific and hardware-specific
          parameters. For setting parameters, the keyword argument
          syntax should be used, and multiple parameters can be set at once. For
          querying, a parameter name should be quoted as a string, and only one
          parameter can be queried at a time::
          
           # Set WiFi access point name (formally known as ESSID) and WiFi channel
           ap.config(essid='My AP', channel=11)
           # Query params one by one
           print(ap.config('essid'))
           print(ap.config('channel'))
      """

class WLAN:
    """
   This class provides a driver for WiFi network processors.  Example usage::
   
       import network
       # enable station interface and connect to WiFi access point
       nic = network.WLAN(network.STA_IF)
       nic.active(True)
       nic.connect('your-ssid', 'your-password')
       # now use sockets as usual
   """

    def __init__(self, interface_id: int, /):
        """
      Create a WLAN network interface object. Supported interfaces are
      ``network.STA_IF`` (station aka client, connects to upstream WiFi access
      points) and ``network.AP_IF`` (access point, allows other WiFi clients to
      connect). Availability of the methods below depends on interface type.
      For example, only STA interface may `WLAN.connect()` to an access point.
      """
    @overload
    def active(self, /) -> bool:
        """
       Activate ("up") or deactivate ("down") network interface, if boolean
       argument is passed. Otherwise, query current state if no argument is
       provided. Most other methods require active interface.
      """
    @overload
    def active(self, is_active: bool, /) -> None:
        """
       Activate ("up") or deactivate ("down") network interface, if boolean
       argument is passed. Otherwise, query current state if no argument is
       provided. Most other methods require active interface.
      """
    def connect(
        self,
        ssid: str | None = None,
        password: str | None = None,
        /,
        *,
        bssid: bytes | None = None,
    ) -> None:
        """
       Connect to the specified wireless network, using the specified password.
       If *bssid* is given then the connection will be restricted to the
       access-point with that MAC address (the *ssid* must also be specified
       in this case).
      """
    def disconnect(self) -> None:
        """
       Disconnect from the currently connected wireless network.
      """
    def scan(self) -> tuple[str, bytes, int, int, int]:
        """
       Scan for the available wireless networks.
       Hidden networks -- where the SSID is not broadcast -- will also be scanned
       if the WLAN interface allows it.
       
       Scanning is only possible on STA interface. Returns list of tuples with
       the information about WiFi access points:
       
           (ssid, bssid, channel, RSSI, authmode, hidden)
       
       *bssid* is hardware address of an access point, in binary form, returned as
       bytes object. You can use `binascii.hexlify()` to convert it to ASCII form.
       
       There are five values for authmode:
       
           * 0 -- open
           * 1 -- WEP
           * 2 -- WPA-PSK
           * 3 -- WPA2-PSK
           * 4 -- WPA/WPA2-PSK
       
       and two for hidden:
       
           * 0 -- visible
           * 1 -- hidden
      """
    @overload
    def status(self) -> int:
        """
       Return the current status of the wireless connection.
       
       When called with no argument the return value describes the network link status.
       The possible statuses are defined as constants:
       
           * ``STAT_IDLE`` -- no connection and no activity,
           * ``STAT_CONNECTING`` -- connecting in progress,
           * ``STAT_WRONG_PASSWORD`` -- failed due to incorrect password,
           * ``STAT_NO_AP_FOUND`` -- failed because no access point replied,
           * ``STAT_CONNECT_FAIL`` -- failed due to other problems,
           * ``STAT_GOT_IP`` -- connection successful.
       
       When called with one argument *param* should be a string naming the status
       parameter to retrieve.  Supported parameters in WiFI STA mode are: ``'rssi'``.
      """
    @overload
    def status(self, param: str, /) -> int:
        """
       Return the current status of the wireless connection.
       
       When called with no argument the return value describes the network link status.
       The possible statuses are defined as constants:
       
           * ``STAT_IDLE`` -- no connection and no activity,
           * ``STAT_CONNECTING`` -- connecting in progress,
           * ``STAT_WRONG_PASSWORD`` -- failed due to incorrect password,
           * ``STAT_NO_AP_FOUND`` -- failed because no access point replied,
           * ``STAT_CONNECT_FAIL`` -- failed due to other problems,
           * ``STAT_GOT_IP`` -- connection successful.
       
       When called with one argument *param* should be a string naming the status
       parameter to retrieve.  Supported parameters in WiFI STA mode are: ``'rssi'``.
      """
    def isconnected(self) -> bool:
        """
       In case of STA mode, returns ``True`` if connected to a WiFi access
       point and has a valid IP address.  In AP mode returns ``True`` when a
       station is connected. Returns ``False`` otherwise.
      """
    @overload
    def ifconfig(self) -> tuple[str, str, str, str]:
        """
      Get/set IP-level network interface parameters: IP address, subnet mask,
      gateway and DNS server. When called with no arguments, this method returns
      a 4-tuple with the above information. To set the above values, pass a
      4-tuple with the required information.  For example::
      
       nic.ifconfig(('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    def ifconfig(self, ip_mask_gateway_dns: tuple[str, str, str, str], /) -> None:
        """
      Get/set IP-level network interface parameters: IP address, subnet mask,
      gateway and DNS server. When called with no arguments, this method returns
      a 4-tuple with the above information. To set the above values, pass a
      4-tuple with the required information.  For example::
      
       nic.ifconfig(('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    def config(self, param: str, /) -> Any:
        """
      Get or set general network interface parameters. These methods allow to work
      with additional parameters beyond standard IP configuration (as dealt with by
      `WLAN.ifconfig()`). These include network-specific and hardware-specific
      parameters. For setting parameters, keyword argument syntax should be used,
      multiple parameters can be set at once. For querying, parameters name should
      be quoted as a string, and only one parameter can be queries at time::
      
       # Set WiFi access point name (formally known as ESSID) and WiFi channel
       ap.config(essid='My AP', channel=11)
       # Query params one by one
       print(ap.config('essid'))
       print(ap.config('channel'))
      
      Following are commonly supported parameters (availability of a specific parameter
      depends on network technology type, driver, and :term:`MicroPython port`).
      
      =============  ===========
      Parameter      Description
      =============  ===========
      mac            MAC address (bytes)
      essid          WiFi access point name (string)
      channel        WiFi channel (integer)
      hidden         Whether ESSID is hidden (boolean)
      authmode       Authentication mode supported (enumeration, see module constants)
      password       Access password (string)
      dhcp_hostname  The DHCP hostname to use
      reconnects     Number of reconnect attempts to make (integer, 0=none, -1=unlimited)
      =============  ===========
      """
    @overload
    def config(self, **kwargs: Any) -> None:
        """
      Get or set general network interface parameters. These methods allow to work
      with additional parameters beyond standard IP configuration (as dealt with by
      `WLAN.ifconfig()`). These include network-specific and hardware-specific
      parameters. For setting parameters, keyword argument syntax should be used,
      multiple parameters can be set at once. For querying, parameters name should
      be quoted as a string, and only one parameter can be queries at time::
      
       # Set WiFi access point name (formally known as ESSID) and WiFi channel
       ap.config(essid='My AP', channel=11)
       # Query params one by one
       print(ap.config('essid'))
       print(ap.config('channel'))
      
      Following are commonly supported parameters (availability of a specific parameter
      depends on network technology type, driver, and :term:`MicroPython port`).
      
      =============  ===========
      Parameter      Description
      =============  ===========
      mac            MAC address (bytes)
      essid          WiFi access point name (string)
      channel        WiFi channel (integer)
      hidden         Whether ESSID is hidden (boolean)
      authmode       Authentication mode supported (enumeration, see module constants)
      password       Access password (string)
      dhcp_hostname  The DHCP hostname to use
      reconnects     Number of reconnect attempts to make (integer, 0=none, -1=unlimited)
      =============  ===========
      """

class WLANWiPy:
    """
   .. note::
   
       This class is a non-standard WLAN implementation for the WiPy.
       It is available simply as ``network.WLAN`` on the WiPy but is named in the
       documentation below as ``network.WLANWiPy`` to distinguish it from the
       more general :ref:`network.WLAN <network.WLAN>` class.
   
   This class provides a driver for the WiFi network processor in the WiPy. Example usage::
   
       import network
       import time
       # setup as a station
       wlan = network.WLAN(mode=WLAN.STA)
       wlan.connect('your-ssid', auth=(WLAN.WPA2, 'your-key'))
       while not wlan.isconnected():
           time.sleep_ms(50)
       print(wlan.ifconfig())
   
       # now use socket as usual
       ...
   """

    STA: ClassVar[int] = ...
    """
selects the WLAN mode
   """

    AP: ClassVar[int] = ...
    """
selects the WLAN mode
   """

    WEP: ClassVar[int] = ...
    """
selects the network security
   """

    WPA: ClassVar[int] = ...
    """
selects the network security
   """

    WPA2: ClassVar[int] = ...
    """
selects the network security
   """

    INT_ANT: ClassVar[int] = ...
    """
selects the antenna type
   """

    EXT_ANT: ClassVar[int] = ...
    """
selects the antenna type
   """
    @overload
    def __init__(self, id: int = 0, /):
        """
      Create a WLAN object, and optionally configure it. See `init()` for params of configuration.
      
      .. note::
      
      The ``WLAN`` constructor is special in the sense that if no arguments besides the id are given,
      it will return the already existing ``WLAN`` instance without re-configuring it. This is
      because ``WLAN`` is a system feature of the WiPy. If the already existing instance is not
      initialized it will do the same as the other constructors an will initialize it with default
      values.
      """
    @overload
    def __init__(
        self,
        id: int,
        /,
        *,
        mode: int,
        ssid: str,
        auth: tuple[str, str],
        channel: int,
        antenna: int,
    ):
        """
      Create a WLAN object, and optionally configure it. See `init()` for params of configuration.
      
      .. note::
      
      The ``WLAN`` constructor is special in the sense that if no arguments besides the id are given,
      it will return the already existing ``WLAN`` instance without re-configuring it. This is
      because ``WLAN`` is a system feature of the WiPy. If the already existing instance is not
      initialized it will do the same as the other constructors an will initialize it with default
      values.
      """
    def init(
        self,
        mode: int,
        /,
        *,
        ssid: str,
        auth: tuple[str, str],
        channel: int,
        antenna: int,
    ) -> bool:
        """
      Set or get the WiFi network processor configuration.
      
      Arguments are:
      
        - *mode* can be either ``WLAN.STA`` or ``WLAN.AP``.
        - *ssid* is a string with the ssid name. Only needed when mode is ``WLAN.AP``.
        - *auth* is a tuple with (sec, key). Security can be ``None``, ``WLAN.WEP``,
          ``WLAN.WPA`` or ``WLAN.WPA2``. The key is a string with the network password.
          If ``sec`` is ``WLAN.WEP`` the key must be a string representing hexadecimal
          values (e.g. 'ABC1DE45BF'). Only needed when mode is ``WLAN.AP``.
        - *channel* a number in the range 1-11. Only needed when mode is ``WLAN.AP``.
        - *antenna* selects between the internal and the external antenna. Can be either
          ``WLAN.INT_ANT`` or ``WLAN.EXT_ANT``.
      
      For example, you can do::
      
         # create and configure as an access point
         wlan.init(mode=WLAN.AP, ssid='wipy-wlan', auth=(WLAN.WPA2,'www.wipy.io'), channel=7, antenna=WLAN.INT_ANT)
      
      or::
      
         # configure as an station
         wlan.init(mode=WLAN.STA)
      """
    def connect(
        self,
        ssid: str,
        /,
        *,
        auth: tuple[str, str] | None = None,
        bssid: bytes | None = None,
        timeout: int | None = None,
    ) -> None:
        """
      Connect to a WiFi access point using the given SSID, and other security
      parameters.
      
         - *auth* is a tuple with (sec, key). Security can be ``None``, ``WLAN.WEP``,
           ``WLAN.WPA`` or ``WLAN.WPA2``. The key is a string with the network password.
           If ``sec`` is ``WLAN.WEP`` the key must be a string representing hexadecimal
           values (e.g. 'ABC1DE45BF').
         - *bssid* is the MAC address of the AP to connect to. Useful when there are several
           APs with the same ssid.
         - *timeout* is the maximum time in milliseconds to wait for the connection to succeed.
      """
    def scan(self) -> tuple[str, bytes, int, int | None, int]:
        """
      Performs a network scan and returns a list of named tuples with (ssid, bssid, sec, channel, rssi).
      Note that channel is always ``None`` since this info is not provided by the WiPy.
      """
    def disconnect(self) -> None:
        """
      Disconnect from the WiFi access point.
      """
    def isconnected(self) -> bool:
        """
      In case of STA mode, returns ``True`` if connected to a WiFi access point and has a valid IP address.
      In AP mode returns ``True`` when a station is connected, ``False`` otherwise.
      """
    @overload
    def ifconfig(self, if_id: int = 0, /) -> tuple[str, str, str, str]:
        """
      With no parameters given returns a 4-tuple of *(ip, subnet_mask, gateway, DNS_server)*.
      
      if ``'dhcp'`` is passed as a parameter then the DHCP client is enabled and the IP params
      are negotiated with the AP.
      
      If the 4-tuple config is given then a static IP is configured. For instance::
      
         wlan.ifconfig(config=('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    def ifconfig(
        self, if_id: int = 0, /, *, config: str | tuple[str, str, str, str]
    ) -> None:
        """
      With no parameters given returns a 4-tuple of *(ip, subnet_mask, gateway, DNS_server)*.
      
      if ``'dhcp'`` is passed as a parameter then the DHCP client is enabled and the IP params
      are negotiated with the AP.
      
      If the 4-tuple config is given then a static IP is configured. For instance::
      
         wlan.ifconfig(config=('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    def mode(self) -> int:
        """
      Get or set the WLAN mode.
      """
    @overload
    def mode(self, mode: int, /) -> None:
        """
      Get or set the WLAN mode.
      """
    @overload
    def ssid(self) -> str:
        """
      Get or set the SSID when in AP mode.
      """
    @overload
    def ssid(self, ssid: str, /) -> None:
        """
      Get or set the SSID when in AP mode.
      """
    @overload
    def auth(self) -> int:
        """
      Get or set the authentication type when in AP mode.
      """
    @overload
    def auth(self, auth: int, /) -> None:
        """
      Get or set the authentication type when in AP mode.
      """
    @overload
    def channel(self) -> int:
        """
      Get or set the channel (only applicable in AP mode).
      """
    @overload
    def channel(self, channel: int, /) -> None:
        """
      Get or set the channel (only applicable in AP mode).
      """
    @overload
    def antenna(self) -> int:
        """
      Get or set the antenna type (external or internal).
      """
    @overload
    def antenna(self, antenna: int, /) -> None:
        """
      Get or set the antenna type (external or internal).
      """
    @overload
    def mac(self) -> bytes:
        """
      Get or set a 6-byte long bytes object with the MAC address.
      """
    @overload
    def mac(self, mac: bytes, /) -> None:
        """
      Get or set a 6-byte long bytes object with the MAC address.
      """
    def irq(self, *, handler: Callable[[], None], wake: int) -> Any:
        """
       Create a callback to be triggered when a WLAN event occurs during ``machine.SLEEP``
       mode. Events are triggered by socket activity or by WLAN connection/disconnection.
       
           - *handler* is the function that gets called when the IRQ is triggered.
           - *wake* must be ``machine.SLEEP``.
       
       Returns an IRQ object.
      """

class CC3K:
    """
   This class provides a driver for CC3000 WiFi modules.  Example usage::
   
       import network
       nic = network.CC3K(pyb.SPI(2), pyb.Pin.board.Y5, pyb.Pin.board.Y4, pyb.Pin.board.Y3)
       nic.connect('your-ssid', 'your-password')
       while not nic.isconnected():
           pyb.delay(50)
       print(nic.ifconfig())
   
       # now use socket as usual
       ...
   
   For this example to work the CC3000 module must have the following connections:
   
       - MOSI connected to Y8
       - MISO connected to Y7
       - CLK connected to Y6
       - CS connected to Y5
       - VBEN connected to Y4
       - IRQ connected to Y3
   
   It is possible to use other SPI buses and other pins for CS, VBEN and IRQ.
   """

    WEP: ClassVar[int] = ...
    """
security type to use
   """

    WPA: ClassVar[int] = ...
    """
security type to use
   """

    WPA2: ClassVar[int] = ...
    """
security type to use
   """
    def __init__(
        self, spi: pyb.SPI, pin_cs: pyb.Pin, pin_en: pyb.Pin, pin_irq: pyb.Pin, /
    ):
        """
      Create a CC3K driver object, initialise the CC3000 module using the given SPI bus
      and pins, and return the CC3K object.
      
      Arguments are:
      
        - *spi* is an :ref:`SPI object <pyb.SPI>` which is the SPI bus that the CC3000 is
          connected to (the MOSI, MISO and CLK pins).
        - *pin_cs* is a :ref:`Pin object <pyb.Pin>` which is connected to the CC3000 CS pin.
        - *pin_en* is a :ref:`Pin object <pyb.Pin>` which is connected to the CC3000 VBEN pin.
        - *pin_irq* is a :ref:`Pin object <pyb.Pin>` which is connected to the CC3000 IRQ pin.
      
      All of these objects will be initialised by the driver, so there is no need to
      initialise them yourself.  For example, you can use::
      
        nic = network.CC3K(pyb.SPI(2), pyb.Pin.board.Y5, pyb.Pin.board.Y4, pyb.Pin.board.Y3)
      """
    def connect(
        self,
        ssid: str,
        key: str | None = None,
        /,
        *,
        security: int = WPA2,
        bssid: bytes | None = None,
    ) -> None:
        """
      Connect to a WiFi access point using the given SSID, and other security
      parameters.
      """
    def disconnect(self) -> None:
        """
      Disconnect from the WiFi access point.
      """
    def isconnected(self) -> bool:
        """
      Returns True if connected to a WiFi access point and has a valid IP address,
      False otherwise.
      """
    def ifconfig(self) -> tuple[str, str, str, str, str, str, str]:
        """
      Returns a 7-tuple with (ip, subnet mask, gateway, DNS server, DHCP server,
      MAC address, SSID).
      """
    def patch_version(self) -> str:
        """
      Return the version of the patch program (firmware) on the CC3000.
      """
    def patch_program(self, cmd: str, /) -> None:
        """
      Upload the current firmware to the CC3000.  You must pass 'pgm' as the first
      argument in order for the upload to proceed.
      """

class WIZNET5K:
    """
   This class allows you to control WIZnet5x00 Ethernet adaptors based on
   the W5200 and W5500 chipsets.  The particular chipset that is supported
   by the firmware is selected at compile-time via the MICROPY_PY_WIZNET5K
   option.
   
   Example usage::
   
       import network
       nic = network.WIZNET5K(pyb.SPI(1), pyb.Pin.board.X5, pyb.Pin.board.X4)
       print(nic.ifconfig())
   
       # now use socket as usual
       ...
   
   For this example to work the WIZnet5x00 module must have the following connections:
   
       - MOSI connected to X8
       - MISO connected to X7
       - SCLK connected to X6
       - nSS connected to X5
       - nRESET connected to X4
   
   It is possible to use other SPI buses and other pins for nSS and nRESET.
   """

    def __init__(self, spi: pyb.SPI, pin_cs: pyb.Pin, pin_rst: pyb.Pin, /):
        """
      Create a WIZNET5K driver object, initialise the WIZnet5x00 module using the given
      SPI bus and pins, and return the WIZNET5K object.
      
      Arguments are:
      
        - *spi* is an :ref:`SPI object <pyb.SPI>` which is the SPI bus that the WIZnet5x00 is
          connected to (the MOSI, MISO and SCLK pins).
        - *pin_cs* is a :ref:`Pin object <pyb.Pin>` which is connected to the WIZnet5x00 nSS pin.
        - *pin_rst* is a :ref:`Pin object <pyb.Pin>` which is connected to the WIZnet5x00 nRESET pin.
      
      All of these objects will be initialised by the driver, so there is no need to
      initialise them yourself.  For example, you can use::
      
        nic = network.WIZNET5K(pyb.SPI(1), pyb.Pin.board.X5, pyb.Pin.board.X4)
      """
    def isconnected(self) -> bool:
        """
      Returns ``True`` if the physical Ethernet link is connected and up.
      Returns ``False`` otherwise.
      """
    @overload
    def ifconfig(self) -> tuple[str, str, str, str]:
        """
      Get/set IP address, subnet mask, gateway and DNS.
      
      When called with no arguments, this method returns a 4-tuple with the above information.
      
      To set the above values, pass a 4-tuple with the required information.  For example::
      
       nic.ifconfig(('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    @overload
    def ifconfig(self, config: tuple[str, str, str, str], /):
        """
      Get/set IP address, subnet mask, gateway and DNS.
      
      When called with no arguments, this method returns a 4-tuple with the above information.
      
      To set the above values, pass a 4-tuple with the required information.  For example::
      
       nic.ifconfig(('192.168.0.4', '255.255.255.0', '192.168.0.1', '8.8.8.8'))
      """
    def regs(self) -> Any:
        """
      Dump the WIZnet5x00 registers.  Useful for debugging.
      """



"""
Frame buffer manipulation.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/framebuf.rst.

This module provides a general frame buffer which can be used to create
bitmap images, which can then be sent to a display.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import overload, Final

from uio import AnyWritableBuf

MONO_VLSB: Final[int] = ...
"""
Monochrome (1-bit) color format
    This defines a mapping where the bits in a byte are vertically mapped with
    bit 0 being nearest the top of the screen. Consequently each byte occupies
    8 vertical pixels. Subsequent bytes appear at successive horizontal
    locations until the rightmost edge is reached. Further bytes are rendered
    at locations starting at the leftmost edge, 8 pixels lower.
"""

MONO_HLSB: Final[int] = ...
"""
Monochrome (1-bit) color format
    This defines a mapping where the bits in a byte are horizontally mapped.
    Each byte occupies 8 horizontal pixels with bit 7 being the leftmost.
    Subsequent bytes appear at successive horizontal locations until the
    rightmost edge is reached. Further bytes are rendered on the next row, one
    pixel lower.
"""

MONO_HMSB: Final[int] = ...
"""
Monochrome (1-bit) color format
    This defines a mapping where the bits in a byte are horizontally mapped.
    Each byte occupies 8 horizontal pixels with bit 0 being the leftmost.
    Subsequent bytes appear at successive horizontal locations until the
    rightmost edge is reached. Further bytes are rendered on the next row, one
    pixel lower.
"""

RGB565: Final[int] = ...
"""
Red Green Blue (16-bit, 5+6+5) color format
"""

GS2_HMSB: Final[int] = ...
"""
Grayscale (2-bit) color format
"""

GS4_HMSB: Final[int] = ...
"""
Grayscale (4-bit) color format
"""

GS8: Final[int] = ...
"""
Grayscale (8-bit) color format
"""

class FrameBuffer:
    """
   The FrameBuffer class provides a pixel buffer which can be drawn upon with
   pixels, lines, rectangles, text and even other FrameBuffer's. It is useful
   when generating output for displays.
   
   For example::
   
       import framebuf
   
       # FrameBuffer needs 2 bytes for every RGB565 pixel
       fbuf = framebuf.FrameBuffer(bytearray(100 * 10 * 2), 100, 10, framebuf.RGB565)
   
       fbuf.fill(0)
       fbuf.text('MicroPython!', 0, 0, 0xffff)
       fbuf.hline(0, 9, 96, 0xffff)
   """

    def __init__(
        self,
        buffer: AnyWritableBuf,
        width: int,
        height: int,
        format: int,
        stride: int = ...,
        /,
    ):
        """
       Construct a FrameBuffer object.  The parameters are:
       
           - *buffer* is an object with a buffer protocol which must be large
             enough to contain every pixel defined by the width, height and
             format of the FrameBuffer.
           - *width* is the width of the FrameBuffer in pixels
           - *height* is the height of the FrameBuffer in pixels
           - *format* specifies the type of pixel used in the FrameBuffer;
             permissible values are listed under Constants below. These set the
             number of bits used to encode a color value and the layout of these
             bits in *buffer*.
             Where a color value c is passed to a method, c is a small integer
             with an encoding that is dependent on the format of the FrameBuffer.
           - *stride* is the number of pixels between each horizontal line
             of pixels in the FrameBuffer. This defaults to *width* but may
             need adjustments when implementing a FrameBuffer within another
             larger FrameBuffer or screen. The *buffer* size must accommodate
             an increased step size.
       
       One must specify valid *buffer*, *width*, *height*, *format* and
       optionally *stride*.  Invalid *buffer* size or dimensions may lead to
       unexpected errors.
      """
    def fill(self, c: int, /) -> None:
        """
       Fill the entire FrameBuffer with the specified color.
      """
    @overload
    def pixel(self, x: int, y: int, /) -> int:
        """
       If *c* is not given, get the color value of the specified pixel.
       If *c* is given, set the specified pixel to the given color.
      """
    @overload
    def pixel(self, x: int, y: int, c: int, /) -> None:
        """
       If *c* is not given, get the color value of the specified pixel.
       If *c* is given, set the specified pixel to the given color.
      """
    def hline(self, x: int, y: int, w: int, c: int, /) -> None:
        """
       Draw a line from a set of coordinates using the given color and
       a thickness of 1 pixel. The `line` method draws the line up to
       a second set of coordinates whereas the `hline` and `vline`
       methods draw horizontal and vertical lines respectively up to
       a given length.
      """
    def vline(self, x: int, y: int, h: int, c: int, /) -> None:
        """
       Draw a line from a set of coordinates using the given color and
       a thickness of 1 pixel. The `line` method draws the line up to
       a second set of coordinates whereas the `hline` and `vline`
       methods draw horizontal and vertical lines respectively up to
       a given length.
      """
    def line(self, x1: int, y1: int, x2: int, y2: int, c: int, /) -> None:
        """
       Draw a line from a set of coordinates using the given color and
       a thickness of 1 pixel. The `line` method draws the line up to
       a second set of coordinates whereas the `hline` and `vline`
       methods draw horizontal and vertical lines respectively up to
       a given length.
      """
    def rect(self, x: int, y: int, w: int, h: int, c: int, /) -> None:
        """
       Draw a rectangle at the given location, size and color. The `rect`
       method draws only a 1 pixel outline whereas the `fill_rect` method
       draws both the outline and interior.
      """
    def fill_rect(self, x: int, y: int, w: int, h: int, c: int, /) -> None:
        """
       Draw a rectangle at the given location, size and color. The `rect`
       method draws only a 1 pixel outline whereas the `fill_rect` method
       draws both the outline and interior.
      """
    def text(self, s: str, x: int, y: int, c: int = 1, /) -> None:
        """
       Write text to the FrameBuffer using the the coordinates as the upper-left
       corner of the text. The color of the text can be defined by the optional
       argument but is otherwise a default value of 1. All characters have
       dimensions of 8x8 pixels and there is currently no way to change the font.
      """
    def scroll(self, xstep: int, ystep: int, /) -> None:
        """
       Shift the contents of the FrameBuffer by the given vector. This may
       leave a footprint of the previous colors in the FrameBuffer.
      """
    def blit(
        self,
        fbuf: FrameBuffer,
        x: int,
        y: int,
        key: int = -1,
        pallet: FrameBuffer | None = None,
        /,
    ) -> None:
        """
       Draw another FrameBuffer on top of the current one at the given coordinates.
       If *key* is specified then it should be a color integer and the
       corresponding color will be considered transparent: all pixels with that
       color value will not be drawn.
       
       The *palette* argument enables blitting between FrameBuffers with differing
       formats. Typical usage is to render a monochrome or grayscale glyph/icon to
       a color display. The *palette* is a FrameBuffer instance whose format is
       that of the current FrameBuffer. The *palette* height is one pixel and its
       pixel width is the number of colors in the source FrameBuffer. The *palette*
       for an N-bit source needs 2**N pixels; the *palette* for a monochrome source
       would have 2 pixels representing background and foreground colors. The
       application assigns a color to each pixel in the *palette*. The color of the
       current pixel will be that of that *palette* pixel whose x position is the
       color of the corresponding source pixel.
      """



from pybricks.parameters import Button
from typing import List


class Buttons:
    """
    A stub class to represent the buttons member of the EV3Brick class.
    """

    def __init__(self):
        ...

    def pressed(self) -> List[Button]:
        """
        Checks which buttons are currently pressed.

        Returns:
        List of pressed buttons.
        """
        return []



from pybricks.parameters import Port
from typing import Tuple


class EnergyMeter:
    """
    LEGO® MINDSTORMS® Education NXT Energy Meter.

    Args:
        port (Port): Port to which the sensor is connected.
    """

    def __init__(self, port: Port):
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def storage(self) -> int:
        """
        Gets the total available energy stored in the battery.

        Returns:
            Remaining stored energy in joules.
        """
        return 0

    def input(self) -> Tuple[int, int, int]:
        """
        Measures the electrical signals at the input (bottom) side of the energy meter. It measures the voltage applied to it and the current passing through it. The product of these two values is power. This power value is the rate at which the stored energy increases. This power is supplied by an energy source such as the provided solar panel or an externally driven motor.

        Returns:	
            Voltage (millivolts), current (milliamps), and power (milliwatts) measured at the input port.
        """
        return (0, 0, 0)

    def output(self) -> Tuple[int, int, int]:
        """
        Measures the electrical signals at the output (top) side of the energy meter. It measures the voltage applied to the external load and the current passing to it. The product of these two values is power. This power value is the rate at which the stored energy decreases. This power is consumed by the load, such as a light or a motor.

        Returns:
            Voltage (millivolts), current (milliamps), and power (milliwatts) measured at the output port.
        """
        return (0, 0, 0)



def acos():
    pass


def asin():
    pass


def atan():
    pass


def atan2():
    pass


def ceil():
    pass


def copysign():
    pass


def cos():
    pass


def degrees():
    pass


e = 2.71828


def exp():
    pass


def fabs():
    pass


def floor():
    pass


def fmod():
    pass


def frexp():
    pass


def isfinite():
    pass


def isinf():
    pass


def isnan():
    pass


def ldexp():
    pass


def log():
    pass


def modf():
    pass


pi = 3.14159


def pow():
    pass


def radians():
    pass


def sin():
    pass


def sqrt():
    pass


def tan():
    pass


def trunc():
    pass



"""
heap queue algorithm.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/heapq.rst.
====================================

.. module:: heapq
   :synopsis: heap queue algorithm

|see_cpython_module| :mod:`python:heapq`.

This module implements the
`min heap queue algorithm <https://en.wikipedia.org/wiki/Heap_%28data_structure%29>`_.

A heap queue is essentially a list that has its elements stored in such a way
that the first item of the list is always the smallest.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import TypeVar, Any, Final

_T: Final = TypeVar("_T")

def heappush(heap: list[_T], item: _T, /) -> None:
    """
   Push the ``item`` onto the ``heap``.
   """

def heappop(heap: list[_T], /) -> _T:
    """
   Pop the first item from the ``heap``, and return it.  Raise ``IndexError`` if
   ``heap`` is empty.
   
   The returned item will be the smallest item in the ``heap``.
   """

def heapify(x: list[Any], /) -> None:
    """
   Convert the list ``x`` into a heap.  This is an in-place operation.
   """



# -*- coding: utf-8 -*-

import asyncio
import os
from logging import getLogger
from typing import Sequence

from dbus_next.aio import MessageBus
from dbus_next.constants import BusType
from dbus_next.errors import DBusError

logger = getLogger(__name__)

UDISKS2_BUS_NAME = "org.freedesktop.UDisks2"


def list_volumes_sync() -> Sequence[str]:
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(list_volumes())


async def list_volumes() -> Sequence[str]:
    bus = await MessageBus(bus_type=BusType.SYSTEM).connect()

    introspection = None
    with open(
        os.path.join(os.path.dirname(__file__), "dbus", "org.freedesktop.UDisks2.xml"), "r"
    ) as f:
        introspection = f.read()

    object_manager_introspection = None
    with open(
        os.path.join(os.path.dirname(__file__), "dbus", "org.freedesktop.DBus.ObjectManager.xml"),
        "r",
    ) as f:
        object_manager_introspection = f.read()

    proxy_object = bus.get_proxy_object(
        UDISKS2_BUS_NAME, "/org/freedesktop/UDisks2/Manager", introspection
    )

    interface = proxy_object.get_interface("org.freedesktop.UDisks2.Manager")

    block_devices = await interface.call_get_block_devices({})
    logger.debug(f"Block devices: {block_devices}")

    proxy_object = bus.get_proxy_object(
        UDISKS2_BUS_NAME, "/org/freedesktop/UDisks2", object_manager_introspection
    )
    interface = proxy_object.get_interface("org.freedesktop.DBus.ObjectManager")
    managed_objects = await interface.call_get_managed_objects()
    logger.debug(f"Managed objects: {managed_objects}")

    # Find all drives
    drives = []
    for device, values in managed_objects.items():
        type = next(iter(values))
        if type == "org.freedesktop.UDisks2.Drive":
            drives.append(device)
    logger.debug(f"\nDrives: {drives}\n")

    discovered_usb_drives = []
    for a_drive in drives:
        proxy_object = bus.get_proxy_object(UDISKS2_BUS_NAME, a_drive, introspection)
        interface = proxy_object.get_interface("org.freedesktop.UDisks2.Drive")
        connection_bus = await interface.get_connection_bus()
        time_media_detected = await interface.get_time_media_detected()
        if connection_bus == "usb":
            discovered_usb_drives.append(
                {
                    "drive": a_drive,
                    "time_media_detected": time_media_detected,
                }
            )
    logger.debug(f"\nUSB Drives: {discovered_usb_drives}\n")

    if len(discovered_usb_drives) == 0:
        logger.debug("No USB drive found")
        return []

    # Find the block devices associated with each USB drive
    discovered_block_devices = []
    for block_device in block_devices:
        proxy_object = bus.get_proxy_object(UDISKS2_BUS_NAME, block_device, introspection)
        interface = proxy_object.get_interface("org.freedesktop.UDisks2.Block")
        id_usage = await interface.get_id_usage()
        if id_usage != "filesystem":
            continue
        a_drive = await interface.get_drive()
        for dictionary in discovered_usb_drives:
            if dictionary["drive"] == a_drive:
                discovered_block_devices.append(
                    {
                        "block_device": block_device,
                        "time_media_detected": dictionary["time_media_detected"],
                    }
                )
                break
    logger.debug(f"\nDiscovered Block Devices: {discovered_block_devices}\n")

    if len(discovered_block_devices) == 0:
        logger.debug("No block devices found")
        return []

    # In case there are multiple block devices detected, sort by time the device was detected.
    # Most recently detected are placed first.
    discovered_block_devices = sorted(
        discovered_block_devices, key=lambda x: x.get("time_media_detected")
    )
    discovered_block_devices = [i["block_device"] for i in discovered_block_devices]

    discovered_mount_points = []
    for block_device in discovered_block_devices:
        # Make sure that the filesystem is mounted and get the mount point.
        proxy_object = bus.get_proxy_object(UDISKS2_BUS_NAME, block_device, introspection)
        interface = proxy_object.get_interface("org.freedesktop.UDisks2.Filesystem")

        mount_point = None
        try:
            mount_point = await interface.call_mount({})
        except DBusError as error:
            if "is already mounted" not in error.text:
                raise
            mount_points = await interface.get_mount_points()
            # todo Double check that I don't need to account for endianess or other encoding formats here.
            mount_point = mount_points[0].decode("utf-8")
        discovered_mount_points.append(mount_point.rstrip("\x00"))
    logger.debug(f"\nFilesystem Mount Points: {discovered_mount_points}\n")

    return discovered_mount_points



"""
The `ssl` module provides SSL contexts to wrap sockets in.

|see_cpython_module| :mod:`cpython:ssl`.
"""

from __future__ import annotations

import ssl
from typing import Optional, Tuple

import socketpool
from circuitpython_typing import ReadableBuffer, WriteableBuffer

def create_default_context() -> ssl.SSLContext:
    """Return the default SSLContext."""
    ...

class SSLContext:
    """Settings related to SSL that can be applied to a socket by wrapping it.
    This is useful to provide SSL certificates to specific connections
    rather than all of them."""

    def load_cert_chain(self, certfile: str, keyfile: str) -> None:
        """Load a private key and the corresponding certificate.

        The certfile string must be the path to a single file in PEM format
        containing the certificate as well as any number of CA certificates
        needed to establish the certificate's authenticity.  The keyfile string
        must point to a file containing the private key.
        """
    def load_verify_locations(self, cadata: Optional[str] = None) -> None:
        """Load a set of certification authority (CA) certificates used to validate
        other peers' certificates."""
    def set_default_verify_paths(self) -> None:
        """Load a set of default certification authority (CA) certificates."""
    check_hostname: bool
    """Whether to match the peer certificate's hostname."""
    def wrap_socket(
        self,
        sock: socketpool.Socket,
        *,
        server_side: bool = False,
        server_hostname: Optional[str] = None,
    ) -> ssl.SSLSocket:
        """Wraps the socket into a socket-compatible class that handles SSL negotiation.
        The socket must be of type SOCK_STREAM."""

class SSLSocket:
    """Implements TLS security on a subset of `socketpool.Socket` functions. Cannot be created
    directly. Instead, call `wrap_socket` on an existing socket object.

    Provides a subset of CPython's `ssl.SSLSocket` API. It only implements the versions of
    recv that do not allocate bytes objects."""

    def __hash__(self) -> int:
        """Returns a hash for the Socket."""
        ...
    def __enter__(self) -> SSLSocket:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically closes the Socket when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def accept(self) -> Tuple[SSLSocket, Tuple[str, int]]:
        """Accept a connection on a listening socket of type SOCK_STREAM,
        creating a new socket of type SOCK_STREAM.
        Returns a tuple of (new_socket, remote_address)"""
    def bind(self, address: Tuple[str, int]) -> None:
        """Bind a socket to an address

        :param ~tuple address: tuple of (remote_address, remote_port)"""
        ...
    def close(self) -> None:
        """Closes this Socket"""
    def connect(self, address: Tuple[str, int]) -> None:
        """Connect a socket to a remote address

        :param ~tuple address: tuple of (remote_address, remote_port)"""
        ...
    def listen(self, backlog: int) -> None:
        """Set socket to listen for incoming connections

        :param ~int backlog: length of backlog queue for waiting connetions"""
        ...
    def recv_into(self, buffer: WriteableBuffer, bufsize: int) -> int:
        """Reads some bytes from the connected remote address, writing
        into the provided buffer. If bufsize <= len(buffer) is given,
        a maximum of bufsize bytes will be read into the buffer. If no
        valid value is given for bufsize, the default is the length of
        the given buffer.

        Suits sockets of type SOCK_STREAM
        Returns an int of number of bytes read.

        :param bytearray buffer: buffer to receive into
        :param int bufsize: optionally, a maximum number of bytes to read."""
        ...
    def send(self, bytes: ReadableBuffer) -> int:
        """Send some bytes to the connected remote address.
        Suits sockets of type SOCK_STREAM

        :param ~bytes bytes: some bytes to send"""
        ...
    def settimeout(self, value: int) -> None:
        """Set the timeout value for this socket.

        :param ~int value: timeout in seconds.  0 means non-blocking.  None means block indefinitely.
        """
        ...
    def setblocking(self, flag: bool) -> Optional[int]:
        """Set the blocking behaviour of this socket.

        :param ~bool flag: False means non-blocking, True means block indefinitely."""
        ...



"""
MicroPython on the BBC micro:bit comes with a powerful music and
sound module. It’s very easy to generate bleeps and bloops from
the device if you attach a speaker. Use crocodile clips to attach
pin 0 and GND to the positive and negative inputs on the speaker -
it doesn’t matter which way round they are connected to the speaker.

Musical Notation

    An individual note is specified thus:

    NOTE[octave][:duration]

    For example, A1:4 refers to the note “A” in octave 1 that lasts for four ticks
     (a tick is an arbitrary length of time defined by a tempo setting function -
     see below). If the note name R is used then it is treated as a rest (silence).

    Accidentals (flats and sharps) are denoted by the b
    (flat - a lower case b) and # (sharp - a hash symbol).
    For example, Ab is A-flat and C# is C-sharp.

    Note names are case-insensitive.

    The octave and duration parameters are states that carry over to
    subsequent notes until re-specified. The default states are octave = 4
    (containing middle C) and duration = 4 (a crotchet, given the default
    tempo settings - see below).

    For example, if 4 ticks is a crotchet, the following list is crotchet,
    quaver, quaver, crotchet based arpeggio:

    ['c1:4', 'e:2', 'g', 'c2:4']

    The opening of Beethoven’s 5th Symphony would be encoded thus:

    ['r4:2', 'g', 'g', 'g', 'eb:8', 'r:2', 'f', 'f', 'f', 'd:8']

    The definition and scope of an octave conforms to the table
    listed on this page about scientific pitch notation. For example, middle “C” is
    c4' and concert “A” (440) is 'a4'. Octaves start on the note “C”.
    The definition and scope of an octave conforms to the table
    listed on this page about scientific pitch notation. For example, middle
    “C” is 'c4' and concert “A” (440) is 'a4'. Octaves start on the note “C”.


        Built in Melodies

        For the purposes of education and entertainment, the module contains
        several example tunes that are expressed as Python lists. They can be used like this:

     import music
     music.play(music.NYAN)


    All the tunes are either out of copyright, composed by
    Nicholas H.Tollervey and
    released to the public domain or have an unknown composer and are
    covered by a fair (educational) use provision.

    They are:

    DADADADUM - the opening to Beethoven’s 5th Symphony in C minor.
    ENTERTAINER - the opening fragment of Scott Joplin’s
                  Ragtime classic “The Entertainer”.
    PRELUDE - the opening of the first Prelude in C Major of
              J.S.Bach’s 48 Preludes and Fugues.
    ODE - the “Ode to Joy” theme from Beethoven’s 9th Symphony in D minor.
    NYAN - the Nyan Cat theme (http://www.nyan.cat/). The composer is unknown.
           This is fair use for educational porpoises (as they say in New York).
    RINGTONE - something that sounds like a mobile phone ringtone. To be used
               to indicate an incoming message.
    FUNK - a funky bass line for secret agents and criminal masterminds.
    BLUES - a boogie-woogie 12-bar blues walking bass.
    BIRTHDAY - “Happy Birthday to You...” for copyright status see:
                http://www.bbc.co.uk/news/world-us-canada-34332853
    WEDDING - the bridal chorus from Wagner’s opera “Lohengrin”.
    FUNERAL - the “funeral march” otherwise known as Frédéric Chopin’s
            Piano Sonata No. 2 in B♭ minor, Op. 35.
    PUNCHLINE - a fun fragment that signifies a joke has been made.
    PYTHON - John Philip Sousa’s march “Liberty Bell” aka, the theme for
             “Monty Python’s Flying Circus” (after which the Python
            programming language is named).
    BADDY - silent movie era entrance of a baddy.
    CHASE - silent movie era chase scene.
    BA_DING - a short signal to indicate something has happened.
    WAWAWAWAA - a very sad trombone.
    JUMP_UP - for use in a game, indicating upward movement.
    JUMP_DOWN - for use in a game, indicating downward movement.
    POWER_UP - a fanfare to indicate an achievement unlocked.
    POWER_DOWN - a sad fanfare to indicate an achievement lost.

    Example

        Plays a simple tune using the Micropython music module.
        This example requires a speaker/buzzer/headphones connected to P0 and GND.

    from microbit import *
    import music

    # play Prelude in C.
    notes = [
        'c4:1', 'e', 'g', 'c5', 'e5', 'g4', 'c5', 'e5', 'c4', 'e', 'g', 'c5', 'e5', 'g4', 'c5', 'e5',
        'c4', 'd', 'g', 'd5', 'f5', 'g4', 'd5', 'f5', 'c4', 'd', 'g', 'd5', 'f5', 'g4', 'd5', 'f5',
        'b3', 'd4', 'g', 'd5', 'f5', 'g4', 'd5', 'f5', 'b3', 'd4', 'g', 'd5', 'f5', 'g4', 'd5', 'f5',
        'c4', 'e', 'g', 'c5', 'e5', 'g4', 'c5', 'e5', 'c4', 'e', 'g', 'c5', 'e5', 'g4', 'c5', 'e5',
        'c4', 'e', 'a', 'e5', 'a5', 'a4', 'e5', 'a5', 'c4', 'e', 'a', 'e5', 'a5', 'a4', 'e5', 'a5',
        'c4', 'd', 'f#', 'a', 'd5', 'f#4', 'a', 'd5', 'c4', 'd', 'f#', 'a', 'd5', 'f#4', 'a', 'd5',
        'b3', 'd4', 'g', 'd5', 'g5', 'g4', 'd5', 'g5', 'b3', 'd4', 'g', 'd5', 'g5', 'g4', 'd5', 'g5',
        'b3', 'c4', 'e', 'g', 'c5', 'e4', 'g', 'c5', 'b3', 'c4', 'e', 'g', 'c5', 'e4', 'g', 'c5',
        'b3', 'c4', 'e', 'g', 'c5', 'e4', 'g', 'c5', 'b3', 'c4', 'e', 'g', 'c5', 'e4', 'g', 'c5',
        'a3', 'c4', 'e', 'g', 'c5', 'e4', 'g', 'c5', 'a3', 'c4', 'e', 'g', 'c5', 'e4', 'g', 'c5',
        'd3', 'a', 'd4', 'f#', 'c5', 'd4', 'f#', 'c5', 'd3', 'a', 'd4', 'f#', 'c5', 'd4', 'f#', 'c5',
        'g3', 'b', 'd4', 'g', 'b', 'd', 'g', 'b', 'g3', 'b3', 'd4', 'g', 'b', 'd', 'g', 'b'
    ]

    music.play(notes)
    """
from typing import List, Tuple, Union

from .microbit import MicroBitAnalogDigitalPin, pin0

DADADADUM = 0
ENTERTAINER = 1
PRELUDE = 2
ODE = 3
NYAN = 4
RINGTONE = 5
FUNK = 6
BLUES = 7
BIRTHDAY = 8
WEDDING = 9
FUNERAL = 10
PUNCHLINE = 11
PYTHON = 12
BADDY = 12
CHASE = 13
BA_DING = 14
WAWAWAWAA = 15
JUMP_UP = 16
JUMP_DOWN = 17
POWER_UP = 18
POWER_DOWN = 19


def set_tempo(ticks: int = 4, bpm: int = 120) -> None:
    """
    Sets the approximate tempo for playback.

    A number of ticks (expressed as an integer) constitute a beat.
    Each beat is to be played at a certain frequency per minute
    (expressed as the more familiar BPM - beats per minute -
     also as an integer).

    Suggested default values allow the following useful behaviour:

        music.set_tempo() - reset the tempo to default of ticks = 4, bpm = 120
        music.set_tempo(ticks=8) - change the “definition” of a beat
        music.set_tempo(bpm=180) - just change the tempo

    To work out the length of a tick in milliseconds is very simple arithmetic:
    60000/bpm/ticks_per_beat . For the default values that’s
    60000/120/4 = 125 milliseconds or 1 beat = 500 milliseconds.
    """


def get_tempo(self) -> Tuple[int, int]:
    """
    Gets the current tempo as a tuple of integers: (ticks, bpm).
    """


def play(music: Union[str, List[str]],
         pin: MicroBitAnalogDigitalPin = pin0, wait: bool=True, loop: bool=False) -> None:
    """
    Sets the approximate tempo for playback.

    A number of ticks (expressed as an integer) constitute a beat.
    Each beat is to be played at a certain frequency per
    minute (expressed as the more familiar BPM - beats per minute -
    also as an integer).

    Suggested default values allow the following useful behaviour:

        music.set_tempo() - reset the tempo to default of ticks = 4, bpm = 120
        music.set_tempo(ticks=8) - change the “definition” of a beat
        music.set_tempo(bpm=180) - just change the tempo

    To work out the length of a tick in milliseconds is very simple arithmetic:
    60000/bpm/ticks_per_beat . For the default values that’s 60000/120/4 =
    125 milliseconds or 1 beat = 500 milliseconds.
    """


def pitch(frequency: int, len=-1, pin: MicroBitAnalogDigitalPin = pin0,
          wait: bool=True) -> None:
    """

    Plays a pitch at the integer frequency given for the specified
    number of milliseconds. For example, if the frequency is set to 440 and
    the length to 1000 then we hear a standard concert A for one second.

    If wait is set to True, this function is blocking.

    If len is negative the pitch is played continuously until
    either the blocking call is interrupted or, in the case of a background call,
    a new frequency is set or stop is called (see below).
    """


def stop(pin: MicroBitAnalogDigitalPin = pin0) -> None:
    """
    Stops all music playback on a given pin.
    """


def reset() -> None:
    """
    Resets the state of the following attributes in the following way:

            ticks = 4
            bpm = 120
            duration = 4
            octave = 4
    """



from logging import getLogger
from typing import List

from turcar.plugins.micropython import BareMetalMicroPythonConfigPage, BareMetalMicroPythonProxy
from turcar.plugins.micropython.esptool_dialog import try_launch_esptool_dialog

# from turcar.plugins.micropython.esptool_dialog import try_launch_esptool_dialog
from turcar.plugins.micropython.mp_front import add_micropython_backend, get_uart_adapter_vids_pids
from turcar.plugins.micropython.uf2dialog import show_uf2_installer

logger = getLogger(__name__)

VIDS_PIDS_TO_AVOID_IN_ESP_BACKENDS = set()


class ESPProxy(BareMetalMicroPythonProxy):
    @classmethod
    def get_vids_pids_to_avoid(self):
        return VIDS_PIDS_TO_AVOID_IN_ESP_BACKENDS

    def _get_backend_launcher_path(self) -> str:
        import turcar.plugins.esp.esp_back

        return turcar.plugins.esp.esp_back.__file__


class ESP8266Proxy(ESPProxy):
    description = "MicroPython on ESP8266"
    config_page_constructor = "ESP8266"

    @classmethod
    def get_known_usb_vids_pids(cls):
        return get_uart_adapter_vids_pids()


class ESP32Proxy(ESPProxy):
    @classmethod
    def get_known_usb_vids_pids(cls):
        # See https://github.com/espressif/usb-pids for 0x303A
        return get_uart_adapter_vids_pids() | {(0x303A, None)}

    @classmethod
    def _is_potential_port(cls, p):
        lower_desc = p.description.lower()
        return (
            super()._is_potential_port(p) or "m5stack" in lower_desc or "esp32" in lower_desc
        ) and "circuitpython" not in lower_desc


class ESPConfigPage(BareMetalMicroPythonConfigPage):
    def _open_flashing_dialog(self, kind: str) -> None:
        if kind == "esptool":
            try_launch_esptool_dialog(self.winfo_toplevel(), "MicroPython")
        elif kind == "UF2":
            show_uf2_installer(self, firmware_name="MicroPython")
        else:
            raise ValueError(f"Unexpected kind{kind}")

    @property
    def allow_webrepl(self):
        return True


class ESP8266ConfigPage(ESPConfigPage):
    def get_flashing_dialog_kinds(self) -> List[str]:
        return ["esptool"]


class ESP32ConfigPage(ESPConfigPage):
    def get_flashing_dialog_kinds(self) -> List[str]:
        return ["esptool", "UF2"]


def load_plugin():
    add_micropython_backend(
        "ESP32",
        ESP32Proxy,
        "MicroPython (ESP32)",
        ESP32ConfigPage,
        sort_key="35",
    )
    add_micropython_backend(
        "ESP8266", ESP8266Proxy, "MicroPython (ESP8266)", ESP8266ConfigPage, sort_key="36"
    )



from logging import getLogger
from typing import List

from turcar.plugins.micropython import add_micropython_backend
from turcar.plugins.micropython.mp_common import RAW_PASTE_SUBMIT_MODE
from turcar.plugins.micropython.mp_front import (
    BareMetalMicroPythonConfigPage,
    BareMetalMicroPythonProxy,
    get_uart_adapter_vids_pids,
)
from turcar.plugins.micropython.uf2dialog import show_uf2_installer

logger = getLogger(__name__)

VIDS_PIDS_TO_AVOID_IN_RP2040 = set()


class RP2040BackendProxy(BareMetalMicroPythonProxy):
    @classmethod
    def should_consider_unknown_devices(cls):
        return True

    @classmethod
    def get_known_usb_vids_pids(cls):
        # can be anything
        return set()

    def get_node_label(self):
        return "RP2040 device"

    def _get_backend_launcher_path(self) -> str:
        import turcar.plugins.rp2040.rp2040_back

        return turcar.plugins.rp2040.rp2040_back.__file__

    @classmethod
    def get_vids_pids_to_avoid(self):
        return get_uart_adapter_vids_pids() | VIDS_PIDS_TO_AVOID_IN_RP2040


class RP2040BackendConfigPage(BareMetalMicroPythonConfigPage):
    def get_flashing_dialog_kinds(self) -> List[str]:
        return [""]

    def _open_flashing_dialog(self, kind: str) -> None:
        assert kind == ""
        show_uf2_installer(self, firmware_name="MicroPython")

    @property
    def allow_webrepl(self):
        return True


def load_plugin():
    add_micropython_backend(
        "RP2040",
        RP2040BackendProxy,
        "MicroPython (RP2040)",
        RP2040BackendConfigPage,
        bare_metal=True,
        sort_key="33",
        validate_time=False,
        sync_time=True,
        submit_mode=RAW_PASTE_SUBMIT_MODE,
        write_block_size=64,
    )



"""Support for MIDI synthesis"""

from __future__ import annotations

import typing
from typing import Optional

from circuitpython_typing import ReadableBuffer

def from_file(file: typing.BinaryIO, *, sample_rate: int = 11025) -> MidiTrack:
    """Create an AudioSample from an already opened MIDI file.
    Currently, only single-track MIDI (type 0) is supported.

    :param typing.BinaryIO file: Already opened MIDI file
    :param int sample_rate: The desired playback sample rate; higher sample rate requires more memory


    Playing a MIDI file from flash::

          import audioio
          import board
          import synthio

          data = open("single-track.midi", "rb")
          midi = synthio.from_file(data)
          a = audioio.AudioOut(board.A0)

          print("playing")
          a.play(midi)
          while a.playing:
            pass
          print("stopped")"""
    ...

class MidiTrack:
    """Simple square-wave MIDI synth"""

    def __init__(
        self, buffer: ReadableBuffer, tempo: int, *, sample_rate: int = 11025
    ) -> None:
        """Create a MidiTrack from the given stream of MIDI events. Only "Note On" and "Note Off" events
        are supported; channel numbers and key velocities are ignored. Up to two notes may be on at the
        same time.

        :param ~circuitpython_typing.ReadableBuffer buffer: Stream of MIDI events, as stored in a MIDI file track chunk
        :param int tempo: Tempo of the streamed events, in MIDI ticks per second
        :param int sample_rate: The desired playback sample rate; higher sample rate requires more memory

        Simple melody::

          import audioio
          import board
          import synthio

          dac = audioio.AudioOut(board.SPEAKER)
          melody = synthio.MidiTrack(b"\\0\\x90H\\0*\\x80H\\0\\6\\x90J\\0*\\x80J\\0\\6\\x90L\\0*\\x80L\\0\\6\\x90J\\0" +
                                     b"*\\x80J\\0\\6\\x90H\\0*\\x80H\\0\\6\\x90J\\0*\\x80J\\0\\6\\x90L\\0T\\x80L\\0" +
                                     b"\\x0c\\x90H\\0T\\x80H\\0\\x0c\\x90H\\0T\\x80H\\0", tempo=640)
          dac.play(melody)
          print("playing")
          while dac.playing:
            pass
          print("stopped")"""
        ...
    def deinit(self) -> None:
        """Deinitialises the MidiTrack and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> MidiTrack:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    sample_rate: Optional[int]
    """32 bit value that tells how quickly samples are played in Hertz (cycles per second)."""
